"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

class main:

	def __init__(self, q, limit=3):
		"""
		discord users and servers search

		q          : Query for search
		limit      : Number of pages
		"""

		self.framework = main.framework
		self.q = q
		self._pages = ''
		self.limit = limit
		self._links = []
		self.disboard = 'https://disboard.org/search?'
		self.discordhub = 'https://discordhub.com/user/search?'

	def run_crawl(self):
		query = self.q.split('_')[0]
		url_s = [f"{self.disboard}keyword={query}&page={i}" for i in range(1, self.limit+1)]
		url_u = [f"{self.discordhub}user_search_bar=%23{query}&page={i}" for i in range(1, self.limit+1)]
		if self.q.split('_')[1] == 's':
			max_attempt = len(url_s)
			urls = url_s
		if self.q.split('_')[1] == 'u':
			max_attempt = len(url_u)
			urls = url_u
		for url in range(max_attempt):
			self.framework.verbose(f"[DISCORD] Searching in {url} page...")
			try:
				req = self.framework.request(url=urls[url], allow_redirects=True)
			except Exception as e:
				self.framework.error('ConnectionError', 'util/discord', 'run_crawl')
				max_attempt -= 1
				if max_attempt == 0:
					self.framework.error('Discord is missed!', 'util/discord', 'run_crawl')
					break
			self._pages += req.text
	@property
	def pages(self):
		return self._pages



"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

meta = {
        'name': 'Dictionary',
        'author': 'Kaushik',
        'version': '0.1',
        'description': 'This module uses an unofficial google dictionary api to fetch \
			definitions of query.',
        'sources': ('dictionary',),
        'options': (
                ('query', None, True, 'Query to search', '-q', 'store', str),
        ),
        'examples': ('dictionary -q <QUERY>',)
}

def module_api(self):
	query = self.options['query']
	run = self.dictionary(query)
	run.run_crawl()
	output = {'results': []}
	definitions = run.definitions

	for item in definitions:
		output['results'].append(item)

	self.save_gather(output, 'search/dictionary', query, output=self.options['output'])
	return output

def module_run(self):
	output = module_api(self)['results']
	for item in output:
		print()
		self.output(f"As {item['partOfSpeech']}")
		self.output(item['definition'])



"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

class main:

	def __init__(self, q, limit=1):
		""" searx search engine

			q	: Query for search
			limit	: Number of Pages

		"""
		self.framework = main.framework
		self.q = q
		self._pages = ''
		self.limit = limit
		self._links = []
		self._json = {}
		self.searx = ['https://searx.prvcy.eu/search', 'https://searx.xyz/search', 'https://searx.nevrlands.de/searx/search', 'https://searx.info/search']

	def run_crawl(self):
		page = 1
		self._json['results'] = [] 
		params = {'category_general': 1, 'q': self.q, 
			  'pageno': 1, 'time_range': None, 
			  'format': 'json'}
		max_attempts = 0
		for url in self.searx:
			while True:
				self.framework.verbose(f"[SEARX] Searching {url} page {params['pageno']}...", end='\r')
				try:
					req = self.framework.request(
						url=url,
						params=params,
						allow_redirects= True)
				except Exception as e:
					self.framework.error(f"ConnectionError: {e}", 'util/searx', 'run_crawl')
					max_attempts += 1
					if max_attempts == self.limit:
						self.framework.error(f"Searx {url} is missed!", 'util/searx', 'run_crawl')
						return
				else:
					self._pages += req.text
					if req.status_code == 200 and 'results' in req.json():
						self._json['results'] += req.json()['results']
					else:
						self.framework.error(f"Searx {url} is missed!", 'util/searcx', 'run_crawl')
						break
					if params['pageno'] >= self.limit:
						return
					params['pageno'] += 1

	@property
	def pages(self):
		return self._pages

	@property
	def links(self):
		for result in self._json['results']:
			self._links.append(result['url'])
		return self._links

	@property
	def results(self):
		results = []
		for result in self._json['results']:
			if 'content' in result:
				content = result['content']
			else:
				content = 'No description provided'
			cite = self.framework.meta_search_util().make_cite(result['url'])
			results.append({
				't': result['title'],
				'a': result['url'],
				'c': cite,
				'd': content
			})

		return results
	
	@property
	def json(self):
		return self._json

	@property
	def dns(self):
		return self.framework.page_parse(self._pages).get_dns(self.q)

	@property
	def emails(self):
		return self.framework.page_parse(self._pages).get_emails(self.q)
	
	@property
	def docs(self):
		return self.framework.page_parse(self._pages).get_docs(self.q)



"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import re

class main:

	def __init__(self, url):
		""" Bing mobile friendly test screenshot
		
			url	: url for screenshot
		"""
		self.framework = main.framework
		self.url = url
		self._raw_image_data = []
				
	def screenshot(self):
		self.framework.verbose('[Bing Mobile View] Fetching mobile view of the URL...')
		bing_api_url = 'https://www.bing.com/webmaster/tools/mobile-friendliness-result'
		self.framework._global_options['rand_agent'] = True
		try:
			response = self.framework.request(url=bing_api_url, method='POST', \
			data={'url': self.url, 'retry': '0'}, timeout=40).text # Setting high timeout of as some req take long
		except:
			self.framework.error('ConnectionError.', 'util/bing_mobile_view', 'screenshot')
			return  False
		else:
			
			if 'Bing Webmaster services could not be reached' in response :
				self.framework.error('Bing Webmaster services could not be reached', 'util/bing_mobile_view', 'screenshot')
				return False

			self._raw_image_data = re.findall(r'data\:image.*"', response) # Regex for base64 encoded image 

	@property
	def raw_image_data(self):
		if self._raw_image_data == []: # If no image return empty string else first image
			return ''
		return self._raw_image_data[0]




"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from xml.etree import ElementTree

#Using default empty strings for title and author to avoid author=None and title=None
meta = {
	'name': 'Worldcat Search',
	'author': 'Vikas Kundu',
	'version': '0.1',
	'description': 'Search books on Worldcat using author name or book title. ',
	'sources': ('Worldcat API',),
	'options': (
		('title', '', False, 'Query string for title of a book', '-t', 'store', str),
		('author', '', False, 'Query string for author of a book', '-a', 'store', str),
		('limit', 25, False, 'Search limit(number of results, default=25)', '-l', 'store', int),
		('order', 'desc', False, 'Search order(asc or desc, default=desc)', '-o', 'store', str)
	),
    'examples': ('worldcat -t <TITLE> -l 15',)
}

def module_api(self):
	# Add more error codes when found
	error_codes = {
		'102': 'No data found for this query'
	}
	title = self.options['title']
	author = self.options['author']
	limit = self.options['limit']
	order = self.options['order']
	output = {'results': []}

	run = self.worldcat(title, author, limit, order)
	if run.run_crawl() == False:
		return output
	
	xml_root = ElementTree.fromstring(run.xml_data)	

	#Parsing response code
	response_code = xml_root.findall('.//{http://classify.oclc.org}response')[0].attrib['code']
	if response_code != '4': #If not success
		if response_code in error_codes:
			self.error(f"{error_codes[response_code]}", 'worldcat', 'module_api')
		else:
			self.error('Something went wrong!', 'worldcat', 'module_api')
		return output

	#Parsing total entries available for this query
	workCount = xml_root.findall('.//{http://classify.oclc.org}workCount')
	total_entries = workCount[0].text	
	if limit > int(total_entries): 
		self.alert(f'Only {total_entries} entries available for this query instead of {limit}')
		limit = int(total_entries)

	#Parsing book data
	works = xml_root.findall('.//{http://classify.oclc.org}works')	
	output['results'] = [{} for _ in range(limit)]
	for index in range(0, limit):
    		for i in works:
        		for key,value in i[index].attrib.items():
		            output['results'][index].update({key: value})
	
	self.save_gather(output, 'search/worldcat', title, output=self.options.get('output'))
	return output

def module_run(self):
	output = module_api(self)
	for data in output['results']:
		table_list = [[key, data[key]] for key in data.keys() ]
		self.table(table_list, ['Key', 'Value'], title=f"Entry No. {output['results'].index(data)+1}")



"""
OWASP Maryam!
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

meta = {
	'name': 'Search engine suggestions',
	'author': 'Saeed',
	'version': '0.1',
	'description': 'Keyword autocompleter to find suggestions in search engines',
	'sources': ('google', 'bing', 'yahoo', 'searx', 'peekier', 'gigablast', 'zapmeta', 'millionshort'),
	'comments': (
		"""example: query='google' out: ['google docs', 'google summer of code', 'google maps', 'google mail', 'google news', ..]""",
	),
	'options': (
		('query', None, True, 'keyword, domain name, company name, etc', '-q', 'store', str),
	),
	'examples': ('suggest -q amazon --output',)

}

def module_api(self):
	q = self.options['query']
	run = self.keywords(q)
	run.run_crawl()
	output = {'suggestions': run.keys}
	self.save_gather(output, 'osint/suggest', q, output=self.options['output'])
	return output

def module_run(self):
	self.alert_results(module_api(self))



"""
OWASP Maryam!
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from json import loads

meta = {
	'name': 'Cluster',
	'author': 'Kaushik',
	'version': '0.1',
	'description': 'Cluster your data using kmeans and fp-growth.',
	'required': ('sklearn', 'kneed', 'mlxtend', 'numpy', 'pandas'),
	'options': (
		('json', None, False, 'Json file that contains the data', '-j', 'store', str),
	),
	'examples': ('cluster -j test.json')
}
	
def module_api(self):
	json = self.options['json']
	file = self._is_readable(json)
	if not file:
		return
	data = loads(file.read())
	clusterer = self.cluster(data)
	output = {'json': clusterer.perform_clustering()}
	# self.save_gather(output, 'iris/cluster', json, output=self.options['output'])
	return output

def module_run(self):
	output = module_api(self)['json']
	for index, title in enumerate(output):
		print('\n')
		print(f"CLUSTER {index+1}")
		print(f"TITLE: {title}")
		print('  '+'\n  '.join(output[title]))



"""
OWASP Maryam!
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from inspect import signature

class main:
	def __init__(self, engine_q=None):
		self.framework = main.framework

		if engine_q is None:
			self._engine_q = [
					self.framework.google,
					self.framework.duckduckgo,
					self.framework.bing,
					self.framework.millionshort,
					self.framework.startpage,
					self.framework.dogpile,
					self.framework.qwant,
					self.framework.yandex,
					self.framework.yahoo,
					self.framework.ask,
					self.framework.gigablast,
					self.framework.activesearch
				]
		else:
			self._engine_q = engine_q

		self._error_record = []


	@property
	def _get_new_errors(self):
		new_errors = self.framework._error_stack[len(self._error_record):]
		self._error_record = self.framework._error_stack
		return new_errors

	def search(self,  q, engine=None, limit=1, count=15):	
		if engine is None:
			engine = self._engine_q.pop(0)
		else:
			if engine in self._engine_q:
				self._engine_q.remove(engine)

		results = None

		while results is None:
			sig = signature(engine.__init__)
			if 'limit' in sig.parameters and 'count' in sig.parameters:
				instance = engine(q, limit=limit, count=count)
			elif 'limit' in sig.parameters:
				instance = engine(q, limit=limit)
			elif 'count' in sig.parameters:
				instance = engine(q, count=count)
			else:
				instance = engine(q)

			instance.run_crawl()

			if hasattr(instance,'results'):
				results = instance.results
			else:
				results = instance.links_with_title

			results = results if len(results)>0 else None

			if any('captcha' in x.lower() or 'missed' in x.lower() for x in self.framework._error_stack):
				self.framework._reset_error_stack()
				if len(self._engine_q) != 0:
					engine = self._engine_q.pop(0)
				else:
					self.framework.error('All Engines Exhausted')
					return 

		return results



"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import json

meta = {
	'name': 'Famous Person Report',
	'author': 'Kaushik',
	'version': '0.1',
	'description': 'later',
	'required': ('$search', '$iris'),
	'sources': ('google', 'wikipedia', 'wikileaks', 'twitter', 'sanctionsearch'),
	'options': (
		('name', '', False, 'Name', '-n', 'store', str),
		('family', '', False, 'Family', '-f', 'store', str),
		('query', '', False, 'Query string. Name, Family, or nickname', '-q', 'store', str),
		('limit', 1, False, 'Search limit', '-l', 'store', int),
		('depth', 0, False, 'Search depth (for searching connected people)', '-d', 'store', int),
		('show_plot', False, False, 'show term frequency plot in separate window', '-s', 'store_true', bool),
		('include_tweets', False, False, 'Include recent tweets mentioning the person', '-i', 'store_true', bool),
	),
	'examples': ('famous_person -q <QUERY> -l 5 --output --api',)
}

OUTPUT = {}
VISITED = []

def module_api(self, depth=0):
	global OUPUT, VISITED

	output = {}
	query = self.options['query']
	name = self.options.get('name') or ''
	family = self.options.get('family') or ''
	target_depth = self.options.get('depth') or 0
	show_plot = self.options.get('show_plot')
	include_tweets = self.options.get('include_tweets')

	if name and family:
		fullname = f"{name} {family}"
	elif query:
		fullname = query
	elif name:
		fullname = name
	elif family:
		fullname = family
	else:
		return output

	VISITED.append(fullname)
	if depth != 0:
		print()
		self.output(f"Searching For: {fullname}")

	google_run = self.google(fullname, count=10)
	google_run.run_crawl()
	google_results = google_run.results
	card = google_run.google_card
	wiki = self.wikipedia(fullname, 5)
	wiki.run_crawl()
	links_with_title = wiki.links_with_title

	have_we_wiki = 0
	selected_pid = 0
	selected_link = 0
	selected_title = 0
	# Here I tried to find the best wikipedia page that match the query
	best_match = {}
	score = 0
	for link, title, pid in links_with_title:
		if name in title and family in title and query in title:
			best_match[7+score] = (pid, title, pid)
		elif name in title and family in title:
			best_match[6+score] = (pid, title, pid)
		elif fullname == title:
			best_match[5+score] = (pid, title, pid)
		elif query == title:
			best_match[4+score] = (pid, title, pid)
		elif name.lower() in title and family.lower() in title:
			best_match[3+score] = (pid, title, pid)
		elif fullname in title:
			best_match[2+score] = (pid, title, pid)
		elif query.lower() in title:
			best_match[1+score] = (pid, title, pid)
		score += -1

	if not best_match:
		self.output('We could not find any result in wikipedia.')
	else:
		selected_pid, selected_link, selected_title = best_match[max(best_match.keys())]
		have_we_wiki = 1
		wiki_page = self.wikipedia(selected_pid, 1)
		wiki_extract = wiki_page.page()['extract']
		output['wikipedia'] = f"{selected_link} {selected_title}"

	if 'content' in card:
		output['description'] = card['content'].replace('Description', 'Description: ')
	else:
		if have_we_wiki:
			output['description'] = wiki_extract[:500] 
	if 'name' in card:
		output['name'] = card['name']
	else:
		output['name'] = fullname
	if 'known_as' in card:
		output['known_as'] = card['known_as']
	if 'img' in card:
		output['img'] = card['img']
	else:
		output['img'] = f"https://bing.com/th?q={fullname}"
	if 'info' in card:
		output['info'] = card['info']

	if include_tweets:
		tweets = self.tweet_search(fullname, limit=10)
		tweets.run_crawl()
		output['tweet_mentions'] = tweets.tweets['all']

	output['social'] = card['social']
	output['top_urls'] = []
	keywords = self.keywords(fullname)
	keywords.run_crawl()
	output['top_searched_queries'] = keywords.keys[:15]

	data = ''
	if have_we_wiki:
		data = wiki_extract

	data += ' ' + ' '.join([x['t'] + x['d'] for x in google_results[3:]])
	url_counter = 0
	for i in google_results:
		if url_counter == 3:
			break
		url = i['a']
		if '.wikipedia.org' in url.lower():
			if have_we_wiki:
				continue
		else:
			output['top_urls'].append(url)
		try:
			data += ' ' + self.request(url).text
		except:
			continue
		else:
			url_counter += 1

	self.options['query'] = fullname
	self.options['limit'] = 1
	self.options['id'] = None
	self.options['output'] = False
	sanction = self._loaded_modules['sanctionsearch']
	san_out = sanction.module_api(self)
	if san_out['results']:
		self.options['id'] = int(san_out['results'][0]['link'].split('?id=')[1])
		self.options['query'] = None
		sanction.NAMESEARCH = False
		san_out = sanction.module_api(self)
		output['is_in_blacklist'] = True
		output['usa_blacklist'] = san_out
	else:
		output['is_in_blacklist'] = False
		output['usa_blacklist'] = {}

	self.output('Getting term-frequencies related to person...')
	data +=  ' ' + ' '.join(output['top_searched_queries'])
	plot_histogram = self.tf_histogram(data, 'html')
	plot_histogram.remove_stopwords([name.lower(), family.lower(), fullname.lower(), query.lower()]+query.lower().split(' '))
	output['term_frequency'] = plot_histogram.plot_histogram(f"Top 10 Most Common Words for {fullname}", 15, should_show=show_plot)

	OUTPUT[fullname] = output

	if depth < target_depth:
		child_nodes = []
		for item in output.get('info'):
			if item.startswith('Spouse: '):
				spouses = item[8:].split(',')
				names = list(map(lambda x: x[:x.index("(")].strip(), spouses))
				child_nodes.extend(names)
			if item.startswith('Children: '):
				child_nodes.extend(list(map(lambda x: x.strip(), item[10:].split(','))))

			child_nodes = list(filter(lambda x: 'MORE' not in x, child_nodes))

		for node in child_nodes:
			if node not in VISITED:
				self.options['query'] = node
				module_api(self, depth+1)

def module_run(self):
	module_api(self)
	print()
	self.alert_results(OUTPUT)



"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

meta = {
	'name': 'Photon Search',
	'author': 'Divya Goswami',
	'version': '1.0',
	'description': 'Search-as-you-type and typo tolerant geolocator. Works as reverse geocoder too.',
	'sources': ('photon.komoot.io',),
	'options': (
		('query', '', False, 'Enter a location name', '-q', 'store', str),
		('latitude', '', False, 'Enter latitude', '-lat', 'store', str),
		('longitude', '', False, 'Enter longitude', '-lon', 'store', str),
		('limit', 3, False, 'Search limit(number of results, default=not set)', '-l', 'store', int),
		('engine', 'photon', False, 'Engine used is photon', '-e', 'store', str),
		('thread', 2, False, 'The number of engine that run per round(default=2)', '-t', 'store', int),
	),
	'examples': ('photon -q berlin --output', 
				'photon -q berlin -lat 52 -lon 13',
				'photon -lat 10 -lon 52'
				)
	}

def module_api(self):
	global PAGES, JSON
	query = self.options['query']
	lat = self.options['latitude']
	lon = self.options['longitude']
	limit = self.options['limit']
	engine = self.options['engine']
	output = {}
	q_formats = {
		'detail': f"{query}_{lat}_{lon}",
		'reverse': f"q_{lat}_{lon}"
	}
	if query == '':
		query = q_formats['reverse']
	else:
		query = q_formats['detail']
	run = self.photon(query, limit)
	run.run_crawl()
	PAGES = run.pages
	JSON = run.json
	output = JSON
	self.save_gather(output, 'search/photon', query, output=self.options.get('output'))
	return output

def module_run(self):
	output = module_api(self)['features']
	final = []
	for item in range(len(output)):
		self.output(output[item]['properties']['type'].title())
		self.output(f"Coordinates: {output[item]['geometry']['coordinates']}")
		properties = output[item]['properties']
		prop_k = list(properties.keys())
		for x in prop_k:
			if x not in ['osm_id','osm_type','osm_key','osm_value','countrycode','type','extent','postcode']:
				final.append(properties[x])
		self.output(', '.join(set(final)))
		final = []
		print()



"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import json
import re
import os

class main:

	def __init__(self, q, page, headers):
		""" Web application fingerprint for detect apps. like wappalizer
			
			q 		  : Domain name
			page	  : Web page
			headers	  : Web headers
		"""
		self.framework = main.framework
		self.q = q
		self.page = page
		self.headers = headers
		# Wapps json file
		self.wfile = os.path.join(self.framework.data_path, 'wapps.json')

	def _req_parse(self):
		parser = self.framework.page_parse(self.page)

		# tree = fromstring(self.page)
		# Get Script[src] links
		self.scripts = parser.get_jsfiles
		# Get meta tags
		self.meta = {}
		meta = parser.get_metatags
		for attr in meta:
			if 'name' in attr and 'content' in attr:
				self.meta[attr['name']] = attr['content']

	def _has_app(self, app):
		# Search the easiest things first and save the full-text search of the
		# HTML for last

		for regexes in app['url']:
			for regex in regexes:
				if regex.search(self.q):
					return True
				
		# Headers
		for name, regexes in app['headers'].items():
			if name in self.headers:
				content = self.headers[name]
				# print(regexes)
				for regex in regexes:
					if regex.search(content):
						return True

		# Javascript links
		for regexes in app['scripts']:
			if regexes:
				for regex in regexes:
					for script in self.scripts:
						search = regex.search(script)
						if search:
							return search.group()

		# Meta tags
		for name, regex in app['meta'].items():
			if name in self.meta:
				content = self.meta[name]
				if regex.search(content):
					return True

		# Html search
		for regexes in app['html']:
			if regexes:
				for regex in regexes:
					if regex.search(self.page):
						return True

	def _init_pattern(self, patterns):
		outcome = []
		if isinstance(patterns, list):
			for patt in patterns:
				regex, _, rest = patt.partition('\\;')
				try:
					outcome.append(re.compile(regex, re.I))
				except re.error as e:
					outcome.append(re.compile(r'(?!x)x'))
			return outcome
		else:
			try:
				return [re.compile(patterns, re.I)]
			except re.error as e:
				return [re.compile(r'(?!x)x')]

	def _init_wfile(self):
		with open(self.wfile) as wf:
			jload = json.loads(wf.read())
		self.apps = jload['technologies']
		self.categories = jload['categories']
		"""
		Normalize app data, preparing it for the detection phase.
		"""
		for name, app in self.apps.items():
			self._init_app(app)
	
	def _init_app(self, app):
		# Ensure these keys' values are lists
		for key in ['url', 'html', 'scripts', 'implies']:
			try:
				value = app[key]
			except KeyError:
				app[key] = []
			else:
				if not isinstance(value, list):
					app[key] = [value]

		# Ensure these keys exist
		for key in ['headers', 'meta']:
			try:
				value = app[key]
			except KeyError:
				app[key] = {}

		# Ensure the 'meta' key is a dict
		obj = app['meta']
		if not isinstance(obj, dict):
			app['meta'] = {'generator': obj}

		# Ensure keys are lowercase
		for key in ['headers', 'meta']:
			obj = app[key]
			app[key] = {k.lower(): v for k, v in obj.items()}

		"""
		Strip out key:value pairs from the pattern and compile the regular
		expression.
		"""
		for key in ['url', 'html', 'scripts']:
			app[key] = [self._init_pattern(pattern) for pattern in app[key]]

		for key in ['headers', 'meta']:
			obj = app[key]
			for name, pattern in obj.items():
				obj[name] = self._init_pattern(obj[name])

	def _get_implied_apps(self, detected_apps):
		def __get_implied_apps(apps):
			_implied_apps = set()
			for app in apps:
				try:
					_implied_apps.update(set(self.apps[app]['implies']))
				except KeyError:
					pass
			return _implied_apps

		implied_apps = __get_implied_apps(detected_apps)
		all_implied_apps = set()

		# Descend recursively until we've found all implied apps
		while not all_implied_apps.issuperset(implied_apps):
			all_implied_apps.update(implied_apps)
			implied_apps = __get_implied_apps(all_implied_apps)

		return all_implied_apps

	def get_categories(self, app_name):
		"""
		Returns a list of the categories for an app name.
		"""
		cat_nums = self.apps.get(app_name, {}).get('cats', [])
		cat_names = [self.categories.get(f'{cat_num}', '')
					 for cat_num in cat_nums]

		return cat_names

	def analyze(self):

		detected_apps = set()
		for app_name, app in self.apps.items():
			"""
			Determine whether the web page matches the app signature.
			"""
			has_app = self._has_app(app)
			if has_app:
				detected_apps.add(app_name)

		"""
		Get the set of apps implied by detected_apps
		"""
		detected_apps |= self._get_implied_apps(detected_apps)
		return detected_apps

	def run_crawl(self):
		self._init_wfile()
		self._req_parse()
		app_list = list(self.analyze())
		resp = {}
		for app in app_list:
			app = re.split(r"\\?;", app)[0]
			resp[app] = self.apps[app].get('website')
 
		return resp



"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
# Sources: https://github.com/EnableSecurity/wafw00f and ....

from re import search, I

class main:

	def __init__(self, req):
		""" detect WAFs(Web Application Firewalls)
			
			req		  : Request attribute
		"""
		self.req = req
		self.content = req.text
		self.headers = req.headers
		self.status_code = req.status_code
		self.reason = req.reason
		self._waf = []

	def run_crawl(self):
		for attr in dir(self):
			con1 = not attr.startswith('__')
			con2 = not attr.endswith('__')
			con3 = attr not in ('_waf', 'content', 'headers',
							 'waf', 'run_crawl', 'status_code', 'reason',
							 'header_check', 'status_check', 'reason_check', 'content_check',
							 'cookie_check', 'req', 'framework')
			if con1 and con2 and con3:
				getattr(self, attr)()

	def header_check(self, tup):
		header, match = tup
		value = self.headers.get(header)
		if value:
			# set-cookie can have multiple headers, python gives it to us
			# concatinated with a comma
			if header == 'Set-Cookie':
				values = value.split(', ')
			else:
				values = [value]
			for value in values:
				if search(match, value, I):
					return True
		return False

	def status_check(self, code):
		if self.status_code == code:
			return True
		return False

	def cookie_check(self, match, attack=False):
		return self.header_check(('Set-Cookie', match))

	def reason_check(self, reason_code):
		if str(self.reason) == reason_code:
			return True
		return False

	def content_check(self, regex):
		if search(regex, self.content, I):
			return True
		return False

	def AzionCDN(self):
		s = [
			self.header_check(('Server', r'Azion([-_]CDN)?'))
		]
		if any(s):
			self._waf.append('AzionCDN (AzionCDN)')

	def WatchGuard(self):
		s = [
			self.header_check(('Server', 'WatchGuard')),
			self.content_check(r"Request denied by WatchGuard Firewall"),
			self.content_check(r'WatchGuard Technologies Inc\.')
		]
		if any(s):
			self._waf.append('WatchGuard (WatchGuard Technologies)')

	def iFinity_DotNetNuke(self):
		schema1 = [
			self.header_check(('X-UrlMaster-Debug', '.+')),
			self.header_check(('X-UrlMaster-Ex', '.+')),
		]
		schema2 = [
			self.content_check(r"Ur[li]RewriteModule"),
			self.content_check(r'SecurityCheck')
		]
		if any(i for i in schema1):
			self._waf.append('URLMaster SecurityCheck (iFinity/DotNetNuke)')
		if all(i for i in schema2):
			self._waf.append('URLMaster SecurityCheck (iFinity/DotNetNuke)')

	def PT_Application_Firewall(self):
		s = [
			self.content_check(r'<h1.{0,10}?Forbidden'),
			self.content_check(r'<pre>Request.ID:.{0,10}?\d{4}\-(\d{2})+.{0,35}?pre>')
		]
		if all(i for i in s):
			self._waf.append('PT Application Firewall (Positive Technologies)')
	

	def Indusface(self):
		s = [
			self.header_check(('Server', r'IF_WAF')),
			self.content_check(r'This website is secured against online attacks. Your request was blocked')
		]
		if any(s):
			self._waf.append('IndusGuard (Indusface)')

	def Armor(self):
		s = [
			self.content_check(r'blocked by website protection from armor'),
			self.content_check(r'please create an armor support ticket')
		]
		if any(s):
			self._waf.append('Armor Defense (Armor)')

	def SquidProxy(self):
		s = [
			self.header_check(('Server', r'squid(/[0-9\.]+)?')),
			self.content_check(r'Access control configuration prevents your request')
			]
		if all(i for i in s):
			self._waf.append('SquidProxy IDS (SquidProxy)')

	def ASPA_Firewall(self):
		s = [
			self.header_check(('Server', r'ASPA[\-_]?WAF')),
			self.header_check(('ASPA-Cache-Status', r'.+?'))
		]
		if any(s):
			self._waf.append('ASPA Firewall (ASPA Engineering Co.)')

	def AQTRONIX(self):
		schema1 = [
			self.status_check(999),
			self.reason_check('No Hacking')
		]    
		schema2 = [
			self.status_check(404),
			self.reason_check('Hack Not Found')
		]
		schema3 = [
			self.content_check(r'WebKnight Application Firewall Alert'),
			self.content_check(r'What is webknight\?'),
			self.content_check(r'AQTRONIX WebKnight is an application firewall'),
			self.content_check(r'WebKnight will take over and protect'),
			self.content_check(r'aqtronix\.com/WebKnight'),
			self.content_check(r'AQTRONIX.{0,10}?WebKnight'),
		]
		if all(i for i in schema1):
			self._waf.append('WebKnight (AQTRONIX)')
		if all(i for i in schema2):
			self._waf.append('WebKnight (AQTRONIX)')    
		if any(i for i in schema3):
			self._waf.append('WebKnight (AQTRONIX)')

	def SecureSphere(self):
		s = [
			self.content_check(r'<(title|h2)>Error'),
			self.content_check(r'The incident ID is'),
			self.content_check(r"This page can't be displayed"),
			self.content_check(r'Contact support for additional information')
		]
		if all(i for i in s):
			self._waf.append('SecureSphere (Imperva Inc.)')

	def WP_Cerber_Security(self):
		s = [
			self.content_check(r'your request looks suspicious or similar to automated'),
			self.content_check(r'our server stopped processing your request'),
			self.content_check(r'We.re sorry.{0,10}?you are not allowed to proceed'),
			self.content_check(r'requests from spam posting software'),
			self.content_check(r'<title>403 Access Forbidden')
			]
		if all(i for i in s):
			self._waf.append('WP Cerber Security (Cerber Tech)')

	def Cloudrity(self):
		s = [
			self.content_check(r"Access Denied.{0,10}?Viettel WAF"),
			self.content_check(r"cloudrity\.com\.(vn)?/"),
			self.content_check(r"Viettel WAF System")
		]
		if any(s):
			self._waf.append('Viettel (Cloudrity)')

	def Cloudfloor(self):
		s = [
			self.header_check(('Server', r'CloudfloorDNS(.WAF)?')),
			self.content_check(r'<(title|h\d{1})>CloudfloorDNS.{0,6}?Web Application Firewall Error'),
			self.content_check(r'www\.cloudfloordns\.com/contact')
		]
		if any(s):
			self._waf.append('Cloudfloor (Cloudfloor DNS)')

	def NAXSI(self):
		s = [
			self.header_check(('X-Data-Origin', r'^naxsi(.+)?')),
			self.header_check(('Server', r'naxsi(.+)?')),
			self.content_check(r'blocked by naxsi'),
			self.content_check(r'naxsi blocked information')
		]
		if any(s):
			self._waf.append('NAXSI (NBS Systems)')

	def Distil(self):
		s = [
			self.content_check(r'cdn\.distilnetworks\.com/images/anomaly\.detected\.png'),
			self.content_check(r'distilCaptchaForm'),
			self.content_check(r'distilCallbackGuard')
		]
		if any(s):
			self._waf.append('Distil (Distil Networks)')

	def Edgecast(self):
		s = [
			self.header_check(('Server', r'^ECD(.+)?')),
			self.header_check(('Server', r'^ECS(.*)?'))
		]
		if any(s):
			self._waf.append('Edgecast (Verizon Digital Media)')

	def WebTotem(self):
		s = [
			self.content_check(r"The current request was blocked.{0,8}?>WebTotem")
		]
		if any(s):
			self._waf.append('WebTotem (WebTotem)')

	def Fortinet(self):
		schema1 = [
			self.cookie_check(r'^FORTIWAFSID='),
			self.content_check('.fgd_icon')
		]
		schema2 = [
			self.content_check('fgd_icon'),
			self.content_check('web.page.blocked'),
			self.content_check('url'),
			self.content_check('attack.id'),
			self.content_check('message.id'),
			self.content_check('client.ip')
		]
		if any(i for i in schema1):
			self._waf.append('FortiWeb (Fortinet)')
		if all(i for i in schema2):
			self._waf.append('FortiWeb (Fortinet)')

	def NexusGuard(self):
		s = [
			self.content_check(r'Powered by Nexusguard'),
			self.content_check(r'nexusguard\.com/wafpage/.+#\d{3};')
		]
		if any(s):
			self._waf.append('NexusGuard Firewall (NexusGuard)')

	def Yunsuo(self):
		s = [
			self.cookie_check(r'^yunsuo_session='),
			self.content_check(r'class=\"yunsuologo\"')
		]
		if any(s):
			self._waf.append('Yunsuo (Yunsuo)')

	def PentaWAF(self):
		s = [
			self.header_check(('Server', r'PentaWaf(/[0-9\.]+)?')),
			self.content_check(r'Penta.?Waf/[0-9\.]+?.server')
		]
		if any(s):
			self._waf.append('PentaWAF (Global Network Services)')

	def Anquanbao(self):
		s = [
			self.header_check(('X-Powered-By-Anquanbao', '.+?')),
			self.content_check(r'aqb_cc/error/')
			]
		if any(s):
			self._waf.append('Anquanbao (Anquanbao)')

	def Xuanwudun(self):
		s = [
			self.content_check(r"admin\.dbappwaf\.cn/(index\.php/Admin/ClientMisinform/)?"),
			self.content_check(r'class=.(db[\-_]?)?waf(.)?([\-_]?row)?>')
		]
		if any(s):
			self._waf.append('Xuanwudun (Xuanwudun)')

	def Neusoft(self):
		s = [
			self.content_check(r'SENGINX\-ROBOT\-MITIGATION')
		]
		if any(s):
			self._waf.append('SEnginx (Neusoft)')

	def WebLand(self):
		s = [
			self.header_check(('Server', r'protected by webland'))
		]
		if any(s):
			self._waf.append('WebLand (WebLand)')

	def aeSecure(self):
		s = [
			self.header_check(('aeSecure-code', '.+?')),
			self.content_check(r'aesecure_denied\.png')
		]
		if any(s):
			self._waf.append('aeSecure (aeSecure)')

	def Tencent_Cloud_Firewall(self):
		s = [
			self.content_check(r'waf\.tencent\-?cloud\.com/')
		]
		if any(s):
			self._waf.append('Tencent Cloud Firewall (Tencent Technologies)')

	def VirusDie(self):
		s = [
			self.content_check(r"cdn\.virusdie\.ru/splash/firewallstop\.png"),
			self.content_check(r'copy.{0,10}?Virusdie\.ru')
		]
		if any(s):
			self._waf.append('VirusDie (VirusDie LLC)')

	def YXLink(self):
		s = [
			self.cookie_check(r'^yx_ci_session='),
			self.cookie_check(r'^yx_language='),
			self.header_check(('Server', r'Yxlink([\-_]?WAF)?'))
		]
		if any(s):
			self._waf.append('YXLink (YxLink Technologies)')

	def Amazon(self):
		s = [
			# This is standard detection schema, checking the server header
			self.header_check(('Server', 'Cloudfront')),
			# Found samples returning 'Via: 1.1 58bfg7h6fg76h8fg7jhdf2.cloudfront.net (CloudFront)'
			self.header_check(('Via', r'([0-9\.]+?)? \w+?\.cloudfront\.net \(Cloudfront\)')),
			# The request token is sent along with this header, eg:
			# X-Amz-Cf-Id: sX5QSkbAzSwd-xx3RbJmxYHL3iVNNyXa1UIebDNCshQbHxCjVcWDww==
			self.header_check(('X-Amz-Cf-Id', '.+?')),
			# This is another reliable fingerprint found on headers
			self.header_check(('X-Cache', 'Error from Cloudfront')),
			# These fingerprints are found on the blockpage itself
			self.content_check(r'Generated by cloudfront \(CloudFront\)')
		]
		if any(s):
			self._waf.append('Cloudfront (Amazon)')

	def SpiderLabs(self):
		schema1 = [
			self.header_check(('Server', r'(mod_security|Mod_Security|NOYB)')),
			self.content_check(r'This error was generated by Mod.?Security'),
			self.content_check(r'rules of the mod.security.module'),
			self.content_check(r'mod.security.rules triggered'),
			self.content_check(r'Protected by Mod.?Security'),
			self.content_check(r'/modsecurity[\-_]errorpage/'),
			self.content_check(r'modsecurity iis')
		]
		schema2 = [
			self.reason_check('ModSecurity Action'),
			self.status_check(403)
		]
		schema3 = [
			self.reason_check('ModSecurity Action'),
			self.status_check(406)
		]
		if any(i for i in schema1):
			self._waf.append('ModSecurity (SpiderLabs)')
		if all(i for i in schema2):
			self._waf.append('ModSecurity (SpiderLabs)')
		if all(i for i in schema3):
			self._waf.append('ModSecurity (SpiderLabs)')

	def Amazon(self):
		s = [
			self.header_check(('X-AMZ-ID', '.+?')),
			self.header_check(('X-AMZ-Request-ID', '.+?')),
			self.cookie_check(r'^aws.?alb='),
			self.header_check(('Server', r'aws.?elb'))
		]
		if any(s):
			self._waf.append('AWS Elastic Load Balancer (Amazon)')

	def ChinaCache(self):
		s = [
			self.header_check(('Powered-By-ChinaCache', '.+'))
		]
		if any(s):
			self._waf.append('ChinaCache Load Balancer (ChinaCache)')

	def Cloudflare(self):
		s = [
			self.header_check(('server', 'cloudflare')),
			self.header_check(('server', r'cloudflare[-_]nginx')),
			self.header_check(('cf-ray', r'.+?')),
			self.cookie_check('__cfduid')
		]
		if any(s):
			self._waf.append('Cloudflare (Cloudflare Inc.)')

	def w360WangZhanBao(self):
		s = [
			self.header_check(('Server', r'qianxin\-waf')),
			self.header_check(('WZWS-Ray', r'.+?')),
			self.header_check(('X-Powered-By-360WZB', r'.+?')),
			self.content_check(r'wzws\-waf\-cgi/'),
			self.content_check(r'wangshan\.360\.cn'),
			self.status_check(493)
		]
		if any(s):
			self._waf.append('360WangZhanBao (360 Technologies)')

	def Alert_Logic(self):
		s = [
			self.content_check(r'<(title|h\d{1})>requested url cannot be found'),
			self.content_check(r'we are sorry.{0,10}?but the page you are looking for cannot be found'),
			self.content_check(r'back to previous page'),
			self.content_check(r'proceed to homepage'),
			self.content_check(r'reference id'),
			]
		if all(i for i in s):
			self._waf.append('Alert Logic (Alert Logic)')

	def Janusec(self):
		s = [
			self.content_check(r'janusec application gateway')
		]
		if any(s):
			self._waf.append('Janusec Application Gateway (Janusec)')

	def pkSec(self):
		schema1 = [
			self.content_check(r'pk.?Security.?Module'),
			self.content_check(r'Security.Alert')
		]
		schema2 = [
			self.content_check(r'As this could be a potential hack attack'),
			self.content_check(r'A safety critical (call|request) was (detected|discovered) and blocked'),
			self.content_check(r'maximum number of reloads per minute and prevented access')
		]
		if any(i for i in schema2):
			self._waf.append('pkSecurity IDS (pkSec)')
		if all(i for i in schema1):
			self._waf.append('pkSecurity IDS (pkSec)')

	def PentestIt(self):
		s = [
			self.content_check(r'@?nemesida(\-security)?\.com'),
			self.content_check(r'Suspicious activity detected.{0,10}?Access to the site is blocked'),
			self.content_check(r'nwaf@'),
			self.status_check(222)
		]
		if any(s):
			self._waf.append('Nemesida (PentestIt)')

	def Microsoft(self):
		s = [
			self.content_check(r'The.{0,10}?(isa.)?server.{0,10}?denied the specified uniform resource locator \(url\)'),
		]
		if any(s):
			self._waf.append('ISA Server (Microsoft)')

	def CdnNs_WdidcNet(self):
		s = [
			self.content_check(r'cdnnswaf application gateway')
		]
		if any(s):
			self._waf.append('CdnNS Application Gateway (CdnNs/WdidcNet)')

	def Akamai(self):
		s = [
			self.header_check(('Server', 'AkamaiGHost')),
			self.header_check(('Server', 'AkamaiGHost'))        
		]
		if any(s):
			self._waf.append('Kona SiteDefender (Akamai)')

	def LimeLight(self):
		s = [
			self.cookie_check(r'^limelight'),
			self.cookie_check(r'^l[mg]_sessid=')
		]
		if any(s):
			self._waf.append('LimeLight CDN (LimeLight)')

	def AliYunDun(self):
		s = [
			self.content_check(r'error(s)?\.aliyun(dun)?\.(com|net)?'),
			self.cookie_check(r'^aliyungf_tc='),
			self.content_check(r'cdn\.aliyun(cs)?\.com'),
			self.status_check(405)
			]
		if all(i for i in s):
			self._waf.append('AliYunDun (Alibaba Cloud Computing)')

	def PerimeterX(self):
		s = [
			self.content_check(r'www\.perimeterx\.(com|net)/whywasiblocked'),
			self.content_check(r'client\.perimeterx\.(net|com)'),
			self.content_check(r'denied because we believe you are using automation tools')
		]
		if any(s):
			self._waf.append('PerimeterX (PerimeterX)')

	def RSJoomla(self):
		s = [
			self.content_check(r'com_rsfirewall_(\d{3}_forbidden|event)?')
		]
		if any(s):
			self._waf.append('RSFirewall (RSJoomla!)')

	def Greywizard(self):
		s = [
			self.header_check(('Server', 'greywizard')),
			self.content_check(r'<(title|h\d{1})>Grey Wizard'),
			self.content_check(r'contact the website owner or Grey Wizard'),
			self.content_check(r'We.ve detected attempted attack or non standard traffic from your ip address')
		]
		if any(s):
			self._waf.append('Greywizard (Grey Wizard)')

	def XLabs(self):
		s = [
			self.header_check(('X-CDN', r'XLabs Security')),
			self.header_check(('Secured', r'^By XLabs Security')),
			self.header_check(('Server', r'XLabs[-_]?.?WAF'))
		]
		if any(s):
			self._waf.append('XLabs Security WAF (XLabs)')

	def Incapsula(self):
		s = [
			self.cookie_check(r'^incap_ses.*?='),
			self.cookie_check(r'^visid_incap.*?='),
			self.content_check(r'incapsula incident id'),
			self.content_check(r'powered by incapsula'),
			self.content_check(r'/_Incapsula_Resource')
		]
		if any(s):
			self._waf.append('Incapsula (Imperva Inc.)')

	def Radware(self):
		schema1 = [
			self.content_check(r'CloudWebSec\.radware\.com'),
			self.header_check(('X-SL-CompState', '.+'))
		]
		schema2 = [
			self.content_check(r'because we have detected unauthorized activity'),
			self.content_check(r'<title>Unauthorized Request Blocked'),
			self.content_check(r'if you believe that there has been some mistake'),
			self.content_check(r'\?Subject=Security Page.{0,10}?Case Number')
		]
		if any(i for i in schema1):
			self._waf.append('AppWall (Radware)')
		if all(i for i in schema2):
			self._waf.append('AppWall (Radware)')

	def StackPath(self):
		s = [
			self.content_check(r"This website is using a security service to protect itself"),
			self.content_check(r'You performed an action that triggered the service and blocked your request')
		]
		if all(i for i in s):
			self._waf.append('StackPath (StackPath)')

	def Zenedge(self):
		s = [
			self.header_check(('Server', 'ZENEDGE')),
			self.header_check(('X-Zen-Fury', r'.+?')),
			self.content_check(r'/__zenedge/')
		]
		if any(s):
			self._waf.append('Zenedge (Zenedge)')

	def NewDefend(self):
		s = [
			# This header can be obtained without attack mode
			# Most reliable fingerprint
			self.header_check(('Server', 'Newdefend')),
			# Reliable ones within blockpage
			self.content_check(r'www\.newdefend\.com/feedback'),
			self.content_check(r'/nd\-block/')
		]
		if any(s):
			self._waf.append('Newdefend (NewDefend)')

	def ArmorLogic(self):
		s = [
			self.header_check(('Server', 'Profense')),
			self.cookie_check(r'^PLBSID=')
		]
		if any(s):
			self._waf.append('Profense (ArmorLogic)')

	def OWASP(self):
		s = [
			self.content_check(r'Request rejected by xVarnish\-WAF')
		]
		if any(s):
			self._waf.append('Varnish (OWASP)')

	def Puhui(self):
		s = [
			self.header_check(('Server', r'Puhui[\-_]?WAF'))
		]
		if any(s):
			self._waf.append('Puhui (Puhui)')

	def Sabre(self):
		schema1 = [
			self.content_check(r'dxsupport\.sabre\.com')
		]
		schema2 = [
			self.content_check(r'<title>Application Firewall Error'),
			self.content_check(r'add some important details to the email for us to investigate')
		]
		if any(i for i in schema1):
			self._waf.append('Sabre Firewall (Sabre)')
		if all(i for i in schema2):
			self._waf.append('Sabre Firewall (Sabre)')

	def Microsoft(self):
		s = [
			self.content_check(r'Request Validation has detected a potentially dangerous client input'),
			self.content_check(r'ASP\.NET has detected data in the request'),
			self.content_check(r'HttpRequestValidationException')
		]
		if any(s):
			self._waf.append('RequestValidationMode (Microsoft)')

	def Comodo_cWatch(self):
		s = [
			self.header_check(('Server', r'Protected by COMODO WAF(.+)?'))
		]
		if any(s):
			self._waf.append('Comodo cWatch (Comodo CyberSecurity)')

	def BIG_IP_Local_Traffic_Manager(self):
		s = [
			self.cookie_check('^bigipserver'),
			self.header_check(('X-Cnection', 'close'))
		]
		if any(s):
			self._waf.append('BIG-IP Local Traffic Manager (F5 Networks)')

	def AnYu(self):
		s = [
			self.content_check(r'anyu.{0,10}?the green channel'),
			self.content_check(r'your access has been intercepted by anyu')
			]
		if any(s):
			self._waf.append('AnYu (AnYu Technologies)')

	def GoDaddy(self):
		s = [
			self.content_check(r'GoDaddy (security|website firewall)')
		]
		if any(s):
			self._waf.append('GoDaddy Website Protection (GoDaddy)')

	def Airee(self):
		s = [
			self.header_check(('Server', 'Airee')),
			self.header_check(('X-Cache', r'(\w+\.)?airee\.cloud')),
			self.content_check(r'airee\.cloud')
		]
		if any(s):
			self._waf.append('AireeCDN (Airee)')

	def Barracuda(self):
		s = [
			self.cookie_check(r'^barra_counter_session='),
			self.cookie_check(r'^BNI__BARRACUDA_LB_COOKIE='),
			self.cookie_check(r'^BNI_persistence='),
			self.cookie_check(r'^BN[IE]S_.*?='),
			self.content_check(r'Barracuda.Networks')
		]
		if any(s):
			self._waf.append('Barracuda (Barracuda Networks)')

	def Safe3(self):
		s = [
			self.header_check(('Server', 'Safe3 Web Firewall')),
			self.header_check(('X-Powered-By', r'Safe3WAF/[\.0-9]+?')),
			self.content_check(r'Safe3waf/[0-9\.]+?')
		]
		if any(s):
			self._waf.append('Safe3 Web Firewall (Safe3)')

	def WebARX(self):
		s = [
			self.content_check(r"WebARX.{0,10}?Web Application Firewall"),
			self.content_check(r"www\.webarxsecurity\.com"),
			self.content_check(r'/wp\-content/plugins/webarx/includes/')
		]
		if any(s):
			self._waf.append('WebARX (WebARX Security Solutions)')

	def Yunaq(self):
		s = [
			self.content_check(r'www\.365cyd\.com'),
			self.content_check(r'help\.365cyd\.com/cyd\-error\-help.html\?code=403')
		]
		if any(s):
			self._waf.append('Chuang Yu Shield (Yunaq)')

	def PowerCDN(self):
		s = [
			self.header_check(('Via', r'(.*)?powercdn.com(.*)?')),
			self.header_check(('X-Cache', r'(.*)?powercdn.com(.*)?')),
			self.header_check(('X-CDN', r'PowerCDN'))
		]
		if any(s):
			self._waf.append('PowerCDN (PowerCDN)')

	def Oracle(self):
		s = [
			self.content_check(r'<title>fw_error_www'),
			self.content_check(r'src=\"/oralogo_small\.gif\"'),
			self.content_check(r'www\.oracleimg\.com/us/assets/metrics/ora_ocom\.js')
		]
		if any(s):
			self._waf.append('Oracle Cloud (Oracle)')

	def Barikode(self):
		s = [
			self.content_check(r'<strong>barikode<.strong>'),
		]
		if any(s):
			self._waf.append('Barikode (Ethic Ninja)')

	def UCloud(self):
		s = [
			self.header_check(('Server', r'uewaf(/[0-9\.]+)?')),
			self.content_check(r'/uewaf_deny_pages/default/img/'),
			self.content_check(r'ucloud\.cn')
		]
		if any(s):
			self._waf.append('UEWaf (UCloud)')

	def Eisoo(self):
		s = [
			self.header_check(('Server', r'EisooWAF(\-AZURE)?/?')),
			self.content_check(r'<link.{0,10}?href=\"/eisoo\-firewall\-block\.css'),
			self.content_check(r'www\.eisoo\.com'),
			self.content_check(r'&copy; \d{4} Eisoo Inc')
		]
		if any(s):
			self._waf.append('Eisoo Cloud Firewall (Eisoo)')

	def Teros(self):
		s = [
			self.cookie_check(r'^st8id=')
		]
		if any(s):
			self._waf.append('Teros (Citrix Systems)')

	def CloudLinux(self):
		s = [
			self.header_check(('Server', r'imunify360.{0,10}?')),
			self.content_check(r'protected.by.{0,10}?imunify360'),
			self.content_check(r'powered.by.{0,10}?imunify360'),
			self.content_check(r'imunify360.preloader')
		]
		if any(s):
			self._waf.append('Imunify360 (CloudLinux)')

	def SecKing(self):
		s = [
			self.header_check(('Server', r'secking(.?waf)?'))
		]
		if any(s):
			self._waf.append('SecKing (SecKing)')

	def Mission_Control_Shield(self):
		s = [
			self.header_check(('Server', 'Mission Control Application Shield'))
		]
		if any(s):
			self._waf.append('Mission Control Shield (Mission Control)')

	def BulletProof_Security_Pro(self):
		s = [
			self.content_check(r'\+?bpsMessage'),
			self.content_check(r'403 Forbidden Error Page'),
			self.content_check(r'If you arrived here due to a search')
		]
		if all(i for i in s):
			self._waf.append('BulletProof Security Pro (AITpro Security)')

	def CrawlProtect(self):
		s = [
			self.cookie_check(r'^crawlprotecttag='),
			self.content_check(r'<title>crawlprotect'),
			self.content_check(r'this site is protected by crawlprotect')
		]
		if any(s):
			self._waf.append('CrawlProtect (Jean-Denis Brun)')

	def Sophos(self):
		schema1 = [
			self.content_check(r'www\.sophos\.com'),
			self.content_check(r'Powered by.?(Sophos)? UTM Web Protection')
		]
		schema2 = [
			self.content_check(r'<title>Access to the requested URL was blocked'),
			self.content_check(r'Access to the requested URL was blocked'),
			self.content_check(r'incident was logged with the following log identifier'),
			self.content_check(r'Inbound Anomaly Score exceeded'),
			self.content_check(r'Your cache administrator is')
		]
		if any(i for i in schema1):
			self._waf.append('UTM Web Protection (Sophos)')
		if all(i for i in schema2):
			self._waf.append('UTM Web Protection (Sophos)')

	def Bekchy(self):
		s = [
			# Both signatures are contained within response, so checking for any one of them
			# Sometimes I observed that there is an XHR request being being made to submit the 
			# report data automatically upon page load. In those cases a missing https is causing
			# false negatives.
			self.content_check(r'Bekchy.{0,10}?Access Denied'),
			self.content_check(r'bekchy\.com/report')
		]
		if any(s):
			self._waf.append('Bekchy (Faydata Technologies Inc.)')

	def Beluga(self):
		s = [
			self.header_check(('Server', r'Beluga')),
			self.cookie_check(r'^beluga_request_trail=')
		]
		if any(s):
			self._waf.append('Beluga CDN (Beluga)')

	def NSFocus(self):
		s = [
			self.header_check(('Server', 'NSFocus'))
		]
		if any(s):
			self._waf.append('NSFocus (NSFocus Global Inc.)')

	def FirePass(self):
		schema1 = [
			self.cookie_check('^VHOST'),
			self.header_check(('Location', r'\/my\.logon\.php3'))
		]
		schema2 = [
			self.cookie_check(r'^F5_fire.+?'),
			self.cookie_check('^F5_passid_shrinked')
		]
		if all(i for i in schema1):
			self._waf.append('FirePass (F5 Networks)')
		if all(i for i in schema2):
			self._waf.append('FirePass (F5 Networks)')

	def RayWAF(self):
		s = [
			self.header_check(('Server', r'WebRay\-WAF')),
			self.header_check(('DrivedBy', r'RaySrv.RayEng/[0-9\.]+?'))
		]
		if any(s):
			self._waf.append('RayWAF (WebRay Solutions)')

	def WTS(self):
		s = [
			self.header_check(('Server', r'wts/[0-9\.]+?')),
			self.content_check(r"<(title|h\d{1})>WTS\-WAF")
		]
		if any(s):
			self._waf.append('WTS-WAF (WTS)')

	def Bluedon(self):
		s = [
			# Found sample servers returning 'Server: BDWAF/2.0'
			self.header_check(('Server', r'BDWAF')),
			self.content_check(r'bluedon web application firewall')
		]
		if any(s):
			self._waf.append('Bluedon (Bluedon IST)')

	def BeyondTrust(self):
		s = [
			self.content_check(r'SecureIIS is an internet security application'),
			self.content_check(r'Download SecureIIS Personal Edition'),
			self.content_check(r'https?://www\.eeye\.com/Secure\-?IIS')
		]
		if any(s):
			self._waf.append('eEye SecureIIS (BeyondTrust)')

	def Reblaze(self):
		schema1 = [
			self.cookie_check(r'^rbzid'),
			self.header_check(('Server', 'Reblaze Secure Web Gateway'))
		]
		schema2 = [
			self.content_check(r'current session has been terminated'),
			self.content_check(r'do not hesitate to contact us'),
			self.content_check(r'access denied \(\d{3}\)')
		]
		if any(i for i in schema1):
			self._waf.append('Reblaze (Reblaze)')
		if all(i for i in schema2):
			self._waf.append('Reblaze (Reblaze)')

	def West263CDN(self):
		s = [
			self.header_check(('X-Cache', r'WS?T263CDN'))
		]
		if any(s):
			self._waf.append('West263 CDN (West263CDN)')

	def Shield_Security(self):
		s = [
			self.content_check(r"You were blocked by the Shield"),
			self.content_check(r"remaining transgression\(s\) against this site"),
			self.content_check(r"Something in the URL.{0,5}?Form or Cookie data wasn\'t appropriate")
		]
		if any(s):
			self._waf.append('Shield Security (One Dollar Plugin)')

	def TrueShield(self):
		s = [
			self.content_check(r"SiteLock will remember you"),
			self.content_check(r"Sitelock is leader in Business Website Security Services"),
			self.content_check(r"sitelock[_\-]shield([_\-]logo|[\-_]badge)?"),
			self.content_check(r'SiteLock incident ID')
		]
		if any(s):
			self._waf.append('Sitelock (TrueShield)')

	def Incsub(self):
		schema1 = [
			self.content_check(r'href="http(s)?.\/\/wpmudev.com\/.{0,15}?'),
			self.content_check(r'Click on the Logs tab, then the WAF Log.'),
			self.content_check(r'Choose your site from the list'),
			self.status_check(403)
		]
		schema2 = [
			self.content_check(r'<h1>Whoops, this request has been blocked!'),
			self.content_check(r'This request has been deemed suspicious'),
			self.content_check(r'possible attack on our servers.'),
			self.status_check(403)
		]
		if all(i for i in schema1):
			self._waf.append('wpmudev WAF (Incsub)')
		if all(i for i in schema2):
			self._waf.append('wpmudev WAF (Incsub)')
	

	def Safeline(self):
		s = [
			self.content_check(r'safeline|<!\-\-\sevent id:')
		]
		if any(s):
			self._waf.append('Safeline (Chaitin Tech.)')

	def Microsoft(self):
		s = [
			self.content_check(r"Rejected[-_]By[_-]UrlScan"),
			self.content_check(r'A custom filter or module.{0,4}?such as URLScan')
		]
		if any(s):
			self._waf.append('URLScan (Microsoft)')

	def SiteGround(self):
		s = [
			self.content_check(r"Our system thinks you might be a robot!"),
			self.content_check(r'access is restricted due to a security rule')
		]
		if any(s):
			self._waf.append('SiteGround (SiteGround)')

	def DynamicWeb(self):
		s = [
			self.header_check(('X-403-Status-By', r'dw.inj.check')),
			self.content_check(r'by dynamic check(.{0,10}?module)?')
		]
		if any(s):
			self._waf.append('DynamicWeb Injection Check (DynamicWeb)')

	def DenyALL(self):
		s = [
			self.status_check(200),
			self.reason_check('Condition Intercepted')
		]
		if all(i for i in s):
			self._waf.append('DenyALL (Rohde & Schwarz CyberSecurity)')

	def Microsoft(self):
		s = [
			self.content_check(r'iis (\d+.)+?detailed error'),
			self.content_check(r'potentially dangerous request querystring'),
			self.content_check(r'application error from being viewed remotely (for security reasons)?'),
			self.content_check(r'An application error occurred on the server'),
		]
		if any(s):
			self._waf.append('ASP.NET Generic (Microsoft)')

	def Phion_Ergon(self):
		s = [
			# This method of detection is old (though most reliable), so we check it first
			self.cookie_check(r'^al[_-]?(sess|lb)='),
			self.content_check(r'server detected a syntax error in your request')
			]
		if any(s):
			self._waf.append('Airlock (Phion/Ergon)')

	def CacheFly(self):
		s = [
			self.header_check(('BestCDN', r'Cachefly')),
			self.cookie_check(r'^cfly_req.*=')
		]
		if any(s):
			self._waf.append('CacheFly CDN (CacheFly)')

	def DotDefender(self):
		s = [
			self.header_check(('X-dotDefender-denied', r'.+?')),
			self.content_check(r'dotdefender blocked your request'),
			self.content_check(r'Applicure is the leading provider of web application security')
		]
		if any(s):
			self._waf.append('DotDefender (Applicure Technologies)')

	def Cisco(self):
		s = [
			self.header_check(('Server', 'ACE XML Gateway'))
		]
		if any(s):
			self._waf.append('ACE XML Gateway (Cisco)')

	def NetContinuum(self):
		s = [
			self.cookie_check(r'^NCI__SessionId=')
		]
		if any(s):
			self._waf.append('NetContinuum (Barracuda Networks)')

	def BitNinja(self):
		s = [
			self.content_check(r'Security check by BitNinja'),
			self.content_check(r'Visitor anti-robot validation')
		]
		if any(s):
			self._waf.append('BitNinja (BitNinja)')

	def Sucuri_CloudProxy(self):
		s = [
			self.header_check(('X-Sucuri-ID', r'.+?')),
			self.header_check(('X-Sucuri-Cache', r'.+?')),
			self.header_check(('Server', r'Sucuri(\-Cloudproxy)?')),
			self.header_check(('X-Sucuri-Block', r'.+?')),
			self.content_check(r"Access Denied.{0,6}?Sucuri Website Firewall"),
			self.content_check(r"<title>Sucuri WebSite Firewall.{0,6}?(CloudProxy)?.{0,6}?Access Denied"),
			self.content_check(r"sucuri\.net/privacy\-policy"),
			self.content_check(r"cdn\.sucuri\.net/sucuri[-_]firewall[-_]block\.css"),
			self.content_check(r'cloudproxy@sucuri\.net')
		]
		if any(s):
			self._waf.append('Sucuri CloudProxy (Sucuri Inc.)')

	def Huawei(self):
		s = [
			self.cookie_check(r'^HWWAFSESID='),
			self.header_check(('Server', r'HuaweiCloudWAF')),
			self.content_check(r'hwclouds\.com'),
			self.content_check(r'hws_security@')
		]
		if any(s):
			self._waf.append('Huawei Cloud Firewall (Huawei)')

	def ServerDefender_VP(self):
		s = [
			self.header_check(('X-Pint', r'p(ort\-)?80'))
		]
		if any(s):
			self._waf.append('ServerDefender VP (Port80 Software)')

	def AdNovum(self):
		s = [
			self.cookie_check(r'^Navajo'),
			self.cookie_check(r'^NP_ID')
		]
		if any(s):
			self._waf.append('NevisProxy (AdNovum)')

	def Trafficshield(self):
		s = [
			self.cookie_check('^ASINFO='),
			self.header_check(('Server', 'F5-TrafficShield'))
		]
		if any(s):
			self._waf.append('Trafficshield (F5 Networks)')

	def DOSarrest(self):
		s = [
			self.header_check(('X-DIS-Request-ID', '.+')),
			# Found samples of DOSArrest returning 'Server: DoSArrest/3.5'
			self.header_check(('Server', r'DOSarrest(.*)?'))
		]
		if any(s):
			self._waf.append('DOSarrest (DOSarrest Internet Security)')

	def Astra(self):
		s = [
			self.cookie_check(r'^cz_astra_csrf_cookie'),
			self.content_check(r'astrawebsecurity\.freshdesk\.com'),
			self.content_check(r'www\.getastra\.com/assets/images')
		]
		if any(s):
			self._waf.append('Astra (Czar Securities)')

	def BIG_IP_AP_Manager(self):
		schema1 = [
			self.cookie_check('^LastMRH_Session'),
			self.cookie_check('^MRHSession')
		]
		schema2 = [
			self.cookie_check('^MRHSession'),
			self.header_check(('Server', r'Big([-_])?IP'))
		]
		schema3 = [
			self.cookie_check('^F5_fullWT'),
			self.cookie_check('^F5_fullWT'),
			self.cookie_check('^F5_HT_shrinked')
		]
		if all(i for i in schema1):
			self._waf.append('BIG-IP AP Manager (F5 Networks)')
		if all(i for i in schema2):
			self._waf.append('BIG-IP AP Manager (F5 Networks)')
		if any(i for i in schema3):
			self._waf.append('BIG-IP AP Manager (F5 Networks)')

	def MaxCDN(self):
		s = [
			self.header_check(('X-CDN', r'maxcdn'))
		]
		if any(s):
			self._waf.append('MaxCDN (MaxCDN)')

	def Inactiv(self):
		s = [
			self.content_check(r'firewall.{0,15}?powered.by.{0,15}?malcare.{0,15}?pro'),
			self.content_check('blocked because of malicious activities')
		]
		if any(s):
			self._waf.append('Malcare (Inactiv)')

	def BinarySec(self):
		s = [
			self.header_check(('Server', 'BinarySec')),
			self.header_check(('x-binarysec-via', '.+')),
			self.header_check(('x-binarysec-nocache', '.+'))
		]
		if any(s):
			self._waf.append('BinarySec (BinarySec)')

	def Varnish(self):
		s = [
			self.header_check(('Server', 'Varnish')),
			self.header_check(('X-Varnish', '.+')),
			self.header_check(('X-Cachewall-Action', '.+?')),
			self.header_check(('X-Cachewall-Reason', '.+?')),
			self.content_check(r'security by cachewall'),
			self.content_check(r'403 naughty.{0,10}?not nice!'),
			self.content_check(r'varnish cache server')
		]
		if any(s):
			self._waf.append('CacheWall (Varnish)')

	def FLOSS(self):
		schema1 = [
			self.header_check(('Server', r'^openresty/[0-9\.]+?')),
			self.status_check(403)
		]
		schema2 = [
			self.content_check(r'openresty/[0-9\.]+?'),
			self.status_check(406)
		]
		if all(i for i in schema1):
			self._waf.append('Open-Resty Lua Nginx (FLOSS)')
		if all(i for i in schema2):
			self._waf.append('Open-Resty Lua Nginx (FLOSS)')

	def HyperGuard(self):
		s = [
			self.cookie_check('^WODSESSION=')
		]
		if any(s):
			self._waf.append('HyperGuard (Art of Defense)')

	def Secure_Entry(self):
		s = [
			self.header_check(('Server', 'Secure Entry Server'))
		]
		if any(s):
			self._waf.append('Secure Entry (United Security Providers)')

	def Accenture(self):
		s = [
			self.header_check(('Server', r'ZScaler')),
			self.content_check(r"Access Denied.{0,10}?Accenture Policy"),
			self.content_check(r'policies\.accenture\.com'),
			self.content_check(r'login\.zscloud\.net/img_logo_new1\.png'),
			self.content_check(r'Zscaler to protect you from internet threats'),
			self.content_check(r"Internet Security by ZScaler"),
			self.content_check(r"Accenture.{0,10}?webfilters indicate that the site likely contains")
		]
		if any(s):
			self._waf.append('ZScaler (Accenture)')

	def BlockDoS(self):
		s = [
			self.header_check(('Server', r'blockdos\.net'))
		]
		if any(s):
			self._waf.append('BlockDoS (BlockDoS)')

	def SafeDog(self):
		s = [
			self.cookie_check(r'^safedog\-flow\-item='),
			self.header_check(('Server', 'Safedog')),
			self.content_check(r'safedogsite/broswer_logo\.jpg'),
			self.content_check(r'404\.safedog\.cn/sitedog_stat.html'),
			self.content_check(r'404\.safedog\.cn/images/safedogsite/head\.png')
		]
		if any(s):
			self._waf.append('Safedog (SafeDog)')

	def Yunjiasu(self):
		s = [
			self.header_check(('Server', r'Yunjiasu(.+)?'))
		]
		if any(s):
			self._waf.append('Yunjiasu (Baidu Cloud Computing)')

	def Cloudbric(self):
		s = [
			self.content_check(r'<title>Cloudbric.{0,5}?ERROR!'),
			self.content_check(r'Your request was blocked by Cloudbric'),
			self.content_check(r'please contact Cloudbric Support'),
			self.content_check(r'cloudbric\.zendesk\.com'),
			self.content_check(r'Cloudbric Help Center'),
			self.content_check(r'malformed request syntax.{0,4}?invalid request message framing.{0,4}?or deceptive request routing')
		]
		if any(s):
			self._waf.append('Cloudbric (Penta Security)')

	def Approach(self):
		s = [
			# This method of detection is old (though most reliable), so we check it first
			self.content_check(r'approach.{0,10}?web application (firewall|filtering)'),
			self.content_check(r'approach.{0,10}?infrastructure team')
			]
		if any(s):
			self._waf.append('Approach (Approach)')

	def NinTechNet(self):
		s = [
			self.content_check(r'<title>NinjaFirewall.{0,10}?\d{3}.forbidden'),
			self.content_check(r'For security reasons?.{0,10}?it was blocked and logged')
		]
		if all(i for i in s):
			self._waf.append('NinjaFirewall (NinTechNet)')

	def NetScaler_AppFirewall(self):
		s = [
			# This header can be obtained without attack mode
			self.header_check(('Via', r'NS\-CACHE')),
			# Cookies are set only when someone is authenticated.
			# Not much reliable since wafw00f isn't authenticating.
			self.cookie_check(r'^(ns_af=|citrix_ns_id|NSC_)'),
			self.content_check(r'(NS Transaction|AppFW Session) id'),
			self.content_check(r'Violation Category.{0,5}?APPFW_'),
			self.content_check(r'Citrix\|NetScaler'),
			# Reliable but not all servers return this header
			self.header_check(('Cneonction', r'^(keep alive|close)')),
			self.header_check(('nnCoection', r'^(keep alive|close)'))
		]
		if any(s):
			self._waf.append('NetScaler AppFirewall (Citrix Systems)')

	def NullDDoS(self):
		s = [
			self.header_check(('Server', r'NullDDoS(.System)?'))
		]
		if any(s):
			self._waf.append('NullDDoS Protection (NullDDoS)')

	def KnownSec(self):
		s = [
			self.content_check(r'/ks[-_]waf[-_]error\.png')
		]
		if any(s):
			self._waf.append('KS-WAF (KnownSec)')

	def Defiant(self):
		s = [
			self.header_check(('Server', r'wf[_\-]?WAF')),
			self.content_check(r"Generated by Wordfence"),
			self.content_check(r'broke one of (the )?Wordfence (advanced )?blocking rules'),
			self.content_check(r"/plugins/wordfence")
		]
		if any(s):
			self._waf.append('Wordfence (Defiant)')

	def SecuPress(self):
		s = [
			self.content_check(r'<(title|h\d{1})>SecuPress'),
		]
		if any(s):
			self._waf.append('SecuPress WP Security (SecuPress)')

	def Microsoft(self):
		s = [
			self.header_check(('X-Azure-Ref', '.+?')),
		]
		if any(s):
			self._waf.append('Azure Front Door (Microsoft)')

	def Qcloud(self):
		s = [
			self.content_check(r'腾讯云Web应用防火墙'),
			self.status_check(403)
			]
		if all(i for i in s):
			self._waf.append('Qcloud (Tencent Cloud)')
	

	def IBM(self):
		s = [
			self.header_check(('X-Backside-Transport', r'(OK|FAIL)'))
		]
		if any(s):
			self._waf.append('DataPower (IBM)')

	def TransIP(self):
		s = [
			self.header_check(('X-TransIP-Backend', '.+')),
			self.header_check(('X-TransIP-Balancer', '.+'))
		]
		if any(s):
			self._waf.append('TransIP Web Firewall (TransIP)')

	def ArvanCloud(self):
		s = [
			self.header_check(('Server', 'ArvanCloud'))
		]
		if any(s):
			self._waf.append('ArvanCloud (ArvanCloud)')

	def Dell(self):
		s = [
			self.header_check(('Server', 'SonicWALL')),
			self.content_check(r"<(title|h\d{1})>Web Site Blocked"),
			self.content_check(r'\+?nsa_banner')
		]
		if any(s):
			self._waf.append('SonicWall (Dell)')

	def SiteGuard(self):
		s = [
			self.content_check(r"Powered by SiteGuard"),
			self.content_check(r'The server refuse to browse the page')
		]
		if any(s):
			self._waf.append('SiteGuard (Sakura Inc.)')

	def IBM(self):
		s = [
			self.header_check(('Server', 'WebSEAL')),
			self.content_check(r"This is a WebSEAL error message template file"),
			self.content_check(r"WebSEAL server received an invalid HTTP request")
		]
		if any(s):
			self._waf.append('WebSEAL (IBM)')

	def Zecure(self):
		s = [
			self.content_check(r"<h\d{1}>\d{3}.forbidden<.h\d{1}>"),
			self.content_check(r"request forbidden by administrative rules")
		]
		if all(i for i in s):
			self._waf.append('Shadow Daemon (Zecure)')

	def Instart_DX(self):
		schema1 = [
			self.header_check(('X-Instart-Request-ID', '.+')),
			self.header_check(('X-Instart-Cache', '.+')),
			self.header_check(('X-Instart-WL', '.+'))
		]
		schema2 = [
			self.content_check(r'the requested url was rejected'),
			self.content_check(r'please consult with your administrator'),
			self.content_check(r'your support id is')
		]
		if any(i for i in schema1):
			self._waf.append('Instart DX (Instart Logic)')
		if all(i for i in schema2):
			self._waf.append('Instart DX (Instart Logic)')

	def Qiniu(self):
		s = [
			self.header_check(('X-Qiniu-CDN', r'\d+?'))
		]
		if any(s):
			self._waf.append('Qiniu (Qiniu CDN)')

	def Palo_Alto_Next_Gen_Firewall(self):
		s = [
			self.content_check(r'Download of virus.spyware blocked'),
			self.content_check(r'Palo Alto Next Generation Security Platform')
		]
		if any(s):
			self._waf.append('Palo Alto Next Gen Firewall (Palo Alto Networks)')

	def Fastly(self):
		s = [
			self.header_check(('X-Fastly-Request-ID', r'\w+'))
		]
		if any(s):
			self._waf.append('Fastly (Fastly CDN)')

	def EllisLab(self):
		s = [
			# I have seen some sites use a tracking header and sets a cookie upon authentication
			# 'Set-Cookie: _exp_tracking=rufyhweiuitefgcxyniercyft5-6dctuxeygfr'
			self.cookie_check(r'^exp_track.+?='),
			# There are traces found where cookie is returning values like:
			# Set-Cookie: exp_last_query=834y8d73y94d8g983u4shn8u4shr3uh3
			# Set-Cookie: exp_last_id=b342b432b1a876r8
			self.cookie_check(r'^exp_last_.+?='),
			# In-page fingerprints vary a lot in different sites. Hence these are not quite reliable.
			self.content_check(r'invalid get data')
		]
		if any(s):
			self._waf.append('Expression Engine (EllisLab)')

	def Wallarm(self):
		s = [
			self.header_check(('Server', r'nginx[\-_]wallarm'))
		]
		if any(s):
			self._waf.append('Wallarm (Wallarm Inc.)')

	def KeyCDN(self):
		s = [
			self.header_check(('Server', 'KeyCDN'))
		]
		if any(s):
			self._waf.append('KeyCDN (KeyCDN)')

	def Yundun(self):
		s = [
			self.header_check(('Server', 'YUNDUN')),
			self.header_check(('X-Cache', 'YUNDUN')),
			self.cookie_check(r'^yd_cookie='),
			self.content_check(r'Blocked by YUNDUN Cloud WAF'),
			self.content_check(r'yundun\.com/yd[-_]http[_-]error/'),
			self.content_check(r'www\.yundun\.com/(static/js/fingerprint\d{1}?\.js)?')
		]
		if any(s):
			self._waf.append('Yundun (Yundun)')

	def BlackBaud(self):
		s = [
			self.header_check(('X-Engine', 'onMessage Shield')),
			self.content_check(r'Blackbaud K\-12 conducts routine maintenance'),
			self.content_check(r'onMessage SHEILD'),
			self.content_check(r'maintenance\.blackbaud\.com'),
			self.content_check(r'status\.blackbaud\.com')
		]
		if any(s):
			self._waf.append('OnMessage Shield (BlackBaud)')

	def BIG_IP_AppSec_Manager(self):
		s = [
			self.content_check('the requested url was rejected'),
			self.content_check('please consult with your administrator')
		]
		if all(i for i in s):
			self._waf.append('BIG-IP AppSec Manager (F5 Networks)')

	def LiteSpeed(self):
		schema1 = [
			self.header_check(('Server', 'LiteSpeed')),
			self.status_check(403)
		]
		schema2 = [
			self.content_check(r'Proudly powered by litespeed web server'),
			self.content_check(r'www\.litespeedtech\.com/error\-page')
		]
		if all(i for i in schema1):
			self._waf.append('LiteSpeed (LiteSpeed Technologies)')
		if any(i for i in schema2):
			self._waf.append('LiteSpeed (LiteSpeed Technologies)')

	def Squarespace(self):
		s = [
			self.header_check(('Server', 'Squarespace')),
			self.cookie_check(r'^SS_ANALYTICS_ID='),
			self.cookie_check(r'^SS_MATTR='),
			self.cookie_check(r'^SS_MID='),
			self.cookie_check(r'SS_CVT='),
			self.content_check(r'status\.squarespace\.com'),
			self.content_check(r'BRICK\-\d{2}')
		]
		if any(s):
			self._waf.append('Squarespace (Squarespace)') 

	def Jiasule(self):
		s = [
			self.header_check(('Server', r'jiasule\-waf')),
			self.cookie_check(r'^jsl_tracking(.+)?='),
			self.cookie_check(r'__jsluid='),
			self.content_check(r'notice\-jiasule'),
			self.content_check(r'static\.jiasule\.com')
		]
		if any(s):
			self._waf.append('Jiasule (Jiasule)')



"""
OWASP Maryam!
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


meta = {
	'name': 'Wikipedia Search',
	'author': 'Tarunesh Kumar',
	'version': '0.3',
	'description': 'Search your query in the Wikipedia and show the results.',
	'sources': ('wikipedia',),
	'options': (
		('query', None, False, 'Query string or wiki page id(e.g, 64959383)', '-q', 'store', str),
		('count', 50, False, 'Number of links per page(min=10, max=100, default=50)', '-c', 'store', int),
	),
	'examples': ('wikipedia -q 64959383',)
}

def module_api(self):
	query = self.options['query']
	count = self.options['count']
	wiki = self.wikipedia(query, count)
	target = ''
	output = {'titles': [], 'links': [], 'page': {}}
	if query.isnumeric() and len(query) > 3:
		output['page'] = wiki.page()
	else:
		wiki.run_crawl()
		links_with_title = wiki.links_with_title
		for link, title, pid in links_with_title:
			output['titles'].append(f"{title}[{pid}]")
		output['links'] = wiki.links

	self.save_gather(output, 'search/wikipedia', query, output=self.options.get('output'))	
	return output

def module_run(self):
	output = module_api(self)
	if output['page'] == {}:
		self.alert('links')
		for i in range(len(output['titles'])):
			self.output(output['titles'][i])
			self.output(f"\t{output['links'][i]}", 'G')
		return
	page = {**output['page']}
	target = page.pop('title', self.options['query'])
	if 'extract' not in page:
		self.output('Not found')
		return
	description = page.pop('extract')
	header = ['Key', 'Value']
	self.table(page.items(), header, title=target)
	self.heading('Description')
	self.output(self.textwrapping('\t', description), prep='\t')



"""
OWASP Maryam!
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import concurrent.futures as futures
import copy
from json import loads
from vaderSentiment.vaderSentiment import SentimentIntensityAnalyzer

meta = {
	'name': 'sentiment analysis',
	'author': 'Saeed',
	'version': '0.1',
	'description': 'Running sentiment analysis on your data.',
	'required': ('vaderSentiment',),
	'options': (
		('json', None, False, 'Json file that contains the data', '-j', 'store', str),
		('key', '', False, 'Data key. the value should be a list. None means the json file contains a list: ["..", ..]', '-k', 'store', str),
		('thread', 5, False, 'The number of thread per each sell(default=10)', '-t', 'store', int),
		('pipe', False, False, 'Dev only! pipe data from other modules(default=False)', '-p', 'store_true', bool),
	),
	'examples': ('sentiment -j test.json -k data',
		'sentiment -j test.json -t 50')
}
	
SHOW_MSG = {'neg': 'negative', 'neu': 'neutral', 'pos': 'positive', 'compound': 'compound'}

def dothetask(i, j):
	global OVERALL, SA, MAXES
	for cell in DATA[i:j]:
		p = SA.polarity_scores(cell)
		for i in p:
			OVERALL[i] += p[i]
			if MAXES[i] != []:
				if p[i] > MAXES[i][0]:
					MAXES[i] = [p[i], cell]
			else:
				MAXES[i] = [p[i], cell]

def thread(nthread):
	with futures.ThreadPoolExecutor(max_workers=nthread) as executor:
		tire = round(LEN/nthread)
		if tire > 0:
			for i in range(tire+1):
				mx = i*nthread
				executor.submit(dothetask, mx, mx+nthread)
		else:
			executor.submit(dothetask, 0, LEN)

def module_api(self):
	global DATA, LEN, OVERALL, SA, MAXES
	SA = SentimentIntensityAnalyzer()
	OVERALL = {'neg': 0.0, 'neu': 0.0, 'pos': 0.0, 'compound': 0.0}
	MAXES = {'neg': [], 'neu': [], 'pos': [], 'compound': []}
	if not self.options['pipe']:
		json = self.options['json']
		key = self.options['key']
		file = self._is_readable(json)
		if not file:
			return
		DATA = loads(file.read())
		if key not in DATA:
			self.error("The key doesn't exists", 'module_api', 'iris/sentiment')
			return
		if key != '':
			DATA = DATA[key]
	LEN = len(DATA)
	thread(self.options['thread'])
	if self.options['pipe']:
		return {'overall': OVERALL, 'maxes': MAXES}
	output = {json : {key: [OVERALL, MAXES]}}
	self.save_gather(output, 'osint/docs_search', json, \
		[key], output=self.options['output'])
	return output

def module_run(self):
	module_api(self)
	for i in OVERALL:
		self.alert(SHOW_MSG[i])
		self.output(f"\t{OVERALL[i]}")
		if MAXES[i]:
			self.output(f"most {i} sentence: \t{MAXES[i][1]}")



"""
OWASP Maryam!
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


class main :
    def __init__(self, q, limit=15):
        """ sepiasearch.org search
        
            q 		  : The query for search
            limit	  : The number of details min 15 if exist
        """
        self.q = q
        self.limit = limit
        self.url = 'https://sepiasearch.org/api/v1/search/videos?'
        self.params = {'search': self.q, 'boostLanguages[]': 'en', 'nsfw': 'false', 'start': 0, 'count': self.limit, 'sort': '-match'}
        self.headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        self._collected_data = []
        self._result_found = 0
        
    def run_crawl(self):
        self.framework.verbose('[SEPIASEARCH] Extracting Data From API')
        response = self.framework.request(self.url, params=self.params, headers=self.headers)
        if response.status_code != 200:
            self.framework.error('Request Fail', 'util/sepiasearch', 'run_crawl')
            return
        
        response = response.json()
        self._result_found = response['total']
        for result in response['data']:
            self._collected_data.append({
                'Title': result['name'],
                'Author': result['account']['displayName'],
                'URL': result['url'],
                'Length': self.sec_to_hours(int(result.get('duration'))),
                'Thumbnail': result['thumbnailUrl'],
                'Embeded URL': result.get('embedUrl'),
                'Published Date': result['publishedAt'],
            })

    def sec_to_hours(self, minute):
        """Convert minutes in hours"""
        if isinstance(minute, int):
            minu, sec = divmod(minute, 60)
            return "{}:{}:{} hours".format(*divmod(minu, 60), sec)
        return None

    @property
    def collected_data(self):
        return self._collected_data

    @property
    def total_result_found(self):
        return self._result_found


"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from json import loads

meta = {
	'name': 'Web Applications Identifier',
	'author': 'Saeed',
	'version': '0.3',
	'description': 'Web fingerprinting to identify the applications used with over 1000 pyload.',
	'sources': ('wappalyzer.com', 'github.com/EnableSecurity/wafw00f'),
	'options': (
		('domain', None, True, 'Domain string', '-d', 'store', str),
	),
	'examples': ('wapps -d <DOMAIN>',)
}

def module_api(self):
	domain = self.options['domain']
	req = self.request(domain)
	urlib = self.urlib(domain)
	query = urlib.sub_service('http')
	headers = req.headers
	text = req.text
	resp = {}
	_wafs = self.waf_identify(req)
	_wafs.run_crawl()
	wafs = _wafs._waf
	resp['waf'] = wafs
	v = self.wapps(domain, text, headers)
	wapps = v.run_crawl()
	resp['wapps'] = wapps

	_os = self.os_identify(text, headers)
	_os.run_crawl()
	if _os.os:
		resp['os'] = _os.os

	cms = self.cms_identify(text, headers)
	cms.run_crawl()
	if cms.cms:
		resp['cms'] = cms.cms

	lang = self.lang_identify(text, headers)
	lang.run_crawl()
	if lang.lang:
		resp['lang'] = lang.lang

	self.save_gather(resp, 'footprint/wapps', domain, output=self.options['output'])
	return resp

def module_run(self):
	self.alert_results(module_api(self))



"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

meta = {
    'name': 'Spotify Search',
    'author': 'Kunal Khandelwal',
    'version': '0.1',
    'description': 'Search artists, albums, playlist and users on spotify',
    'sources': ('google', 'yahoo', 'bing','duckduckgo', 'metacrawler', 'millionshort', 'carrot2', 'qwant'),
    'options': (
        ('query', None, True, 'Query string', '-q', 'store', str),
        ('limit', 1, False, 'Search limit(number of pages, default=1)', '-l', 'store', int),
        ('count', 50, False, 'Number of links per page(min=10, max=100, default=50)', '-c', 'store', int),
        ('engine', 'google', False, 'Engine names for search(default=google)', '-e', 'store', str),
        ('thread', 2, False, 'The number of engine that run per round(default=2)', '-t', 'store', int),
    ),
    'examples': ('spotify -q <QUERY> -l 15 --output',)
}

LINKS = []

def search(self, name, q, q_formats, limit, count):
    global LINKS
    engine = getattr(self, name)
    q = q_formats[f"{name}_q"] if f"{name}_q" in q_formats else q_formats['default_q']
    varnames = engine.__init__.__code__.co_varnames
    if 'limit' in varnames and 'count' in varnames:
        attr = engine(q, limit, count)
    elif 'limit' in varnames:
        attr = engine(q, limit)
    else:
        attr = engine(q)

    attr.run_crawl()
    LINKS += attr.links

def module_api(self):
    global LINKS
    query = self.options['query']
    limit = self.options['limit']
    count = self.options['count']
    engines = self.options['engine'].split(',')
    output = {'user': [], 'artist': [], 'playlist': [], 'album': []}
    q_formats = {
        'default_q': f"site:open.spotify.com {query}",
        'millionshort_q': f'site:open.spotify.com "{query}"',
        'qwant_q': f'site:open.spotify.com {query}'
    }

    self.thread(search, self.options['thread'], engines, query, q_formats, limit, count, meta['sources'])

    
    LINKS = list(self.reglib().filter(r"https?://(open\.)?spotify\.com/", list(set(LINKS))))

    for link in self.reglib().filter(r"https?://(open\.)?spotify\.com/user/", LINKS):
        if '/playlist/' not in link:
            output['user'].append(link)
        else:
            output['playlist'].append(link)
    output['artist'] = list(self.reglib().filter(r"https?://(open\.)?spotify\.com/artist/", LINKS))
    output['album'] = list(self.reglib().filter(r"https?://(open\.)?spotify\.com/album/", LINKS))
    output['playlist'] += list(self.reglib().filter(r"https?://(open\.)?spotify\.com/playlist/", LINKS))

    self.save_gather(output, 'search/spotify', query, output=self.options.get('output'))
    return output


def module_run(self):
    self.alert_results(module_api(self))



"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import re
import concurrent.futures

# Web Scraper v5.2

class main:

	def __init__(self, url, debug=False, limit=1, thread_count=1):
		""" Web scraper with thread support

			url		  	 : First page address
			debug	  	 : Show the result at moment
			limit	  	 : Web scraper level(if it's 1 that means just search in first page)
			thread_count : Number of links for each lap
		"""
		self.framework = main.framework
		self.parser = self.framework.urlib(url)
		self.url = self.parser.sub_service(self.framework._global_options['protocol'], ifany=True)
		self.urlib = self.framework.urlib
		self.debug = debug
		self.limit = limit
		self.thread_count = thread_count
		self._CATEGORY_PAGES = {}
		self._PAGES = ''
		self._LINKS = []
		self._OUT_SCOPE_LINKS = []
		self._QUERY_LINKS = []
		self._PHONES = []
		self._CSS = []
		self._JS = []
		self._CDNs = []
		self._MEDIA = []
		self._COMMENTS = []
		self._EMAILS = []
		self._DNS = []
		self._CREDIT_CARDS = []
		self._NETWORKS = {}
		self.numerator = 0
		self.media_exts = ('3gp', '7z', 'aa', 'aac', 'ace', 'aif', 'aiff', 'amr', 'amv', 'amz', 'ape', 'arj', 'asf', 'asf', 'au', 'avi', 'bash', 'bat', 'bin',   
		'bmp', 'bz2', 'c', 'cfa', 'cpp', 'cs', 'csv', 'doc', 'docx', 'f4a', 'f4b', 'f4p', 'f4v', 'flac', 'flv', 'gif', 'gif', 'gifv', 'gz', 
		'gzipico', 'img', 'iso', 'java', 'jfif', 'jpeg', 'jpg', 'lzh', 'm2v', 'm4a', 'm4p', 'm4v', 'md', 'mkv', 'mng', 'mov', 'mp2', 'mp3', 
		'mp4', 'mpa', 'mpe', 'mpeg', 'mpg', 'msi', 'msumpv', 'ogg', 'ogv', 'pdf', 'pl', 'plj', 'png', 'pps', 'ppt', 'ppt', 'pptx', 'pptx', 'py'  
		, 'qt', 'rar', 'rm', 'rm', 'rmvb', 'roq', 'sea', 'sit', 'sitx', 'svg', 'svi', 'tar', 'tar.gz', 'tif', 'tiff', 'tiff', 'vmo', 'vob', '  w64', 
		'wav', 'webm', 'wma', 'wmv', 'wmv', 'woff2', 'wrk', 'wvavi', 'xlsx', 'yaml', 'yml', 'z', 'zip')
		self.passed = []

	# If key not in links append it
	def rept(self, key, _list):
		if isinstance(key, list):
			for i in key:
				i = str(i)
				if i not in _list:
					_list.append(i)
		else:
			if key not in _list:
				_list.append(key)
		return _list

	def debuger(self, val):
		if self.debug:
			self.framework.output(val)

	def add_networks(self, data):
		# Add social nets
		for i in data:
			if i not in self._NETWORKS:
				self._NETWORKS[i] = data[i]
			else:
				self._NETWORKS[i] = self.rept(data[i], self._NETWORKS[i])

	def add_cdn(self, link):
		cond = link[:2] == '//' and '.' in link and link not in self._CDNs
		if cond:
			self.debuger(f'cdn: {link}')
			self.rept(link, self._CDNs)
			return True

	def add_phone(self, link):
		cond = link[:6] == "tel://"
		if cond:
			self.debuger(f'phone: {link}')
			self.rept(link, self._PHONES)
			return True

	def add_email(self, link):
		cond = link.startswith("mailto:")
		if cond:
			self.debuger(f'email: {link[6:]}')
			self.rept(link[6:], self._EMAILS)
			return True

	def joiner(self, url):
		url = url
		urparse = self.urlib(url)
		urparse.url = urparse.quote if '%' not in url else url
		cond1 = url.lower() in ('%20', '', '/', '%23', '#', 'https:', 'http:') or '.' not in url
		cond12 = url.endswith(':')
		cond2 = len(
			urparse.url) > 1 and '%3a//' not in urparse.url and urparse.url[:2] != '//'
		cond3 = urparse.url[:2] == '//'
		if cond1 or cond12:
			return False
		elif cond2:
			urparse.url = self.parser.join(url)
		elif cond3:
			urparse.url = url
		else:
			urparse.url = self.parser.join(url)
		return urparse.url

	def link_category(self, urls):
		links = []
		for url in urls:
			join = self.joiner(url)
			##########################
			# ADD CDN, PHONE and EMAIL
			##########################
			cond1 = not join or (self.add_cdn(url) or self.add_phone(url) or self.add_email(url))
			if cond1:
				continue

			join = str(join).replace('\/', '/')
			##########################
			# ADD OUT SCOPE
			##########################
			urparse = self.urlib(url)
			if urparse.netroot.lower() not in self.url.lower():
				self._OUT_SCOPE_LINKS = self.rept(join, self._OUT_SCOPE_LINKS)
				continue

			##########################
			# ADD QUERY
			##########################
			if urparse.query != '':
				self._QUERY_LINKS = self.rept(join, self._QUERY_LINKS)

			# If the link is a media link(mp4,..) not a web page
			broke = 0
			for ext in self.media_exts:
				if (f'.{ext}/' in join) or join.endswith(f'.{ext}'):
					self._MEDIA = self.rept(join, self._MEDIA)
					broke = 1
					break

			if broke:
				continue
			urlparse2 = self.urlib(join)
			if urparse.check_urlfile('js'):
				self.debuger(f'js: {join}')
				self._JS = self.rept(join, self._JS)
				continue
			elif urparse.check_urlfile('css'):
				self.debuger(f'css: {join}')
				self._CSS = self.rept(join, self._CSS)
				continue
			self._LINKS.append(join)
			links.append(join)
			self.debuger(join)
		return links

	def get_source(self, url):
		self.numerator += 1
		if url in self.passed:
			return []
		self.passed.append(url)
		# Send Request
		try:
			req = self.framework.request(url)
		except:
			return False
		else:
			resp = req.text

		pp = self.framework.page_parse(resp)

		self.add_networks(pp.get_networks)
		self._EMAILS = self.rept(pp.all_emails, self._EMAILS)
		self._COMMENTS = self.rept(pp.get_html_comments, self._COMMENTS)

		# Get all links
		get_links_params = list(set(pp.get_links))

		links = self.link_category(get_links_params)
		self._PAGES += resp
		self._CATEGORY_PAGES[url] = resp
		return links

	def crawl_robots(self):
		self.framework.debug('Crawling robots.txt file...')
		try:
			robots = self.framework.request(f'http://www.{self.urlib(self.url).netroot}/robots.txt')
		except:
			return []
		if robots.status_code == 200:
			links = re.findall(r'Disallow: (.*)?', robots.text)
			links.extend(re.findall(r'Allow: (.*)?', robots.text))
			links.extend(re.findall(r'Sitemap: (.*)?', robots.text))
			makers = []
			for link in links:
				link = self.joiner(link)
				if link:
					makers.append(link)
			return makers
		return []

	def crawl_sitemap(self):
		self.framework.debug('Crawling sitemap.xml file...')
		try:
			sitemap = self.framework.request(f'http://www.{self.urlib(self.url).netroot}/sitemap.xml')
		except:
			return []
		if sitemap.status_code == 200:
			links = re.findall(r'<loc>(.*?)</loc>', sitemap.text) or []
			makers = []
			for link in links:
				if '*' not in link:
					link = self.joiner(link)
					if link:
						makers.append(link)
			return makers
		return []

	# attack function source : https://github.com/s0md3v/Photon
	def attack(self, function, links, thread_count):
		links = list(links)
		threadpool = concurrent.futures.ThreadPoolExecutor(
				max_workers=thread_count)
		futures = (threadpool.submit(function, link) for link in links if link not in self.passed)
		for i, _ in enumerate(concurrent.futures.as_completed(futures)):
			if i + 1 == len(links) or (i + 1) % thread_count == 0:
				print(f'Progress: {i+1}/{len(links)}',
						end='\r')
		print('')

	def run_crawl(self):
		links = self.get_source(self.url)
		if not links:
			return
		if self.limit == 1:
			return
		robmap = []
		self.framework.verbose("Checking robots.txt...")
		robmap.extend(self.crawl_robots())
		self.framework.verbose("Checking sitemap.xml...")
		robmap.extend(self.crawl_sitemap())
		robmap = list(set(links))
		links = self.link_category(robmap)
		links.extend(robmap)
		del robmap

		for depth in range(self.limit):
			if not links:
				break

			print(f"{self.limit} Level {depth+1}: {len(links)} URLs", end='\r')
			try:
				self.attack(self.get_source, list(set(self._LINKS)), self.thread_count)
			except KeyboardInterrupt:
				print('')
				break
		self._LINKS = list(set(self._LINKS))


	@property
	def pages(self):
		return self._PAGES

	@property
	def category_pages(self):
		return self._CATEGORY_PAGES

	@property
	def links(self):
		return self._LINKS

	@property
	def external_links(self):
		return self._OUT_SCOPE_LINKS

	@property
	def query_links(self):
		return self._QUERY_LINKS

	@property
	def js(self):
		return self._JS

	@property
	def css(self):
		return self._CSS

	@property
	def cdn(self):
		return self._CDNs

	@property
	def phones(self):
		return self._PHONES

	@property
	def comments(self):
		return self._COMMENTS

	@property
	def emails(self):
		return self._EMAILS + self.framework.page_parse(self._PAGES.lower()).get_emails(self.parser.netroot.lower())

	@property
	def networks(self):
		return self._NETWORKS

	@property
	def media(self):
		return self._MEDIA



"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import re
from lxml import html

class main:
	def __init__(self, page):
		""" Page parser

			Page  		: web page content
		"""
		self.framework = main.framework
		self.page = page.decode('utf-8') if isinstance(page, bytes) else page

	@property
	def pclean(self):
		subs = r'<em>|<b>|</b>|</em>|<strong>|</strong>|<wbr>|</wbr>|<span class="vivbold qt0">\
				|%22|<span dir="[\w]+">|</span>|<h\d>|</h\d>|<q>|</q>'
		self.remove_comments
		self.page = re.sub(subs, '', self.page)
		self.page = re.sub(r"%3a", ' ', self.page)
		self.page = re.sub(r"%2f", ' ', self.page)
		self.page = re.sub(r"%2f", ' ', self.page)

	def html_fromstring(self, xpath, parent=None, results={}):
		if self.page == '':
			self.framework.error(f"document is nil", 'util/page_parse', 'html_fromstring')
			return False

		if isinstance(xpath, dict):
			for root in xpath:
				results[root] = self.html_fromstring(xpath[root], root)
		elif isinstance(xpath, list):
			results = {}
			for x in xpath:
				results[x] = self.html_fromstring(x, parent)
		else:
			tree = html.fromstring(self.page)
			results = []
			try:
				if parent:
					for node in tree.xpath(parent):
						results += node.xpath(xpath)
				else:
					results = tree.xpath(xpath)
			except:
				self.framework.error(f"invalid xpath: {xpath} or {parent}", 'util/page_parse', 'html_fromstring')
		return results

	def get_engine_results(self, xpaths, xpath_names):
		xpath_results = self.html_fromstring(xpaths)
		results = []
		if xpath_results:
			root = xpath_results[xpath_names['results']]
			for i in range(len(root[xpath_names['results_a']])):
				try:
					result = {
						't': root[xpath_names['results_title']][i].text_content().strip(),
						'a': root[xpath_names['results_a']][i].get('href').strip(),
						'c': root[xpath_names['results_cite']][i].text_content().strip(),
						'd': root[xpath_names['results_content']][i].text_content().strip(),
					}
					results.append(result)
				except:
					pass
			return results
		else:
			return {}

	def dork_clean(self, host):
		# Clear dork's fingerprints
		host = re.sub(r"(['\"]+)|(%40|@)", '', host)
		return host

	def findall(self, reg):
		return re.compile(reg, flags=re.DOTALL).findall(self.page)

	@property
	def sites(self):
		self.pclean
		reg = re.compile(r'<cite>(.*?)</cite>')
		resp = []
		for itr in reg.findall(self.page):
			resp.append(self.framework.urlib(itr).netroot)
		return resp

	@property
	def remove_html_tags(self):
		""" Remove html tags with regex"""
		self.remove_comments
		scripts = re.compile(r"<script[^>]*>.*?</script>", 
				flags=re.DOTALL)
		styles = re.compile(r"<style[^>]*>.*?</style>",
				flags=re.DOTALL)
		tags = re.compile(r'<[^>]+>|&nbsp|&amp|&lt|&gt|&quot|&apos')
		self.page = re.sub(tags, '', re.sub(styles, '', re.sub(scripts, '', self.page)))

	@property
	def remove_comments(self):
		self.page = re.sub(r'(?=<!--)([\s\S]*?)-->', '', self.page)
		self.page = re.sub(r'(?=/\*)([\s\S]*?)\*/', '', self.page)

	@property
	def get_networks(self):
		self.pclean
		reg_id = self.framework.reglib().social_network_ulinks
		resp = {}
		page = self.page.replace('www.', '')
		for i in reg_id:
			if isinstance(reg_id[i], list):
				name = []
				for j in reg_id[i]:
					name += re.findall(j, page)
			else:
				name = re.findall(reg_id[i], page)
			names = []
			for j in name:
				if j not in names:
					names.append(j)
			resp[i] = names
		return resp

	def get_emails(self, host):
		self.pclean
		host = self.dork_clean(host + '.' if '.' not in host else host)
		emails = re.findall(r"[A-z0-9.\-]+@[A-z0-9\-\.]{0,255}?%s" % host, self.page)
		return [x.replace('\\', '') for x in list(set(emails))]

	@property
	def all_emails(self):
		self.pclean
		emails = self.framework.reglib(self.page).emails
		return emails

	def get_dns(self, host, urls=None):
		if urls:
			data = self.dork_clean(str(urls))
		else:
			self.pclean
			data = self.page

		resp = []
		reg = r"[A-z0-9\.\-%s]+\.%s" % ('%', host.replace('"', '').replace("'", ''))
		for i in re.findall(reg, re.sub(r'\\n', '', data)):
			i = i.replace('\\', '').replace('www.', '')
			if i not in resp and '%' not in resp:
				if i.lower().count(host) > 1:
					i = i.split(host)
					for j in i:
						j = f"{j}{host}"
						resp.append(j)
					continue
				resp.append(i)

		return resp

	def get_docs(self, query, urls=None):
		self.pclean
		if '%' in query:
			query = self.framework.urlib(query).unquote
		ext = re.search(r'filetype:([A-z0-9]+)', query)
		if ext:
			docs = []
			ext = f".{ext.group(1)}"
			if urls is None:
				urls = self.get_links
			for url in urls:
				if ext in url:
					docs.append(url)
			return list(set(docs))
		else:
			self.framework.error("Filetype not specified. Concat 'filetype:doc' to the query")
			return []

	def get_attrs(self, tag, attrs=None):
		if attrs is None:
			attrs = []
		reg = r"['\"\s]([^=][\w\-\d_]+)="
		# Find all attribute names
		if not attrs:
			attrs = re.findall(reg, tag)
		resp = {}
		for attr in attrs:
			if not re.search(fr"{attr.lower()}\s*=", tag.lower()):
				continue
			# Get the first char
			fchar = re.search(fr'{attr}=([\s]+)?.', tag)
			content= False
			if fchar:
				c = fchar.group()[-1:]
				# Get content of attrs
				if c not in "'\"":
					content = re.search(fr'{attr}=([^\s>]+)', tag)
				else:
					content = re.search(fr'{attr}={c}([^{c}]*){c}', tag)
			if content:
				content = content.group(1)
			content = content or ''
			resp[attr.lower()] = content
		return resp

	@property
	def get_metatags(self):
		self.remove_comments
		reg = r"<(?i)meta[^>]+/?>"
		reg = re.compile(reg)
		resp = []
		find = reg.findall(self.page)
		for tag in find:
			tag_attrs = self.get_attrs(tag)
			resp.append(tag_attrs)
		return resp

	@property
	def get_jsfiles(self):
		self.remove_comments
		reg = r"<(?i)script[^>]+>"
		reg = re.compile(reg)
		resp = []
		find = reg.findall(self.page)
		for tag in find:
			tag_attrs = self.get_attrs(tag, ['src'])
			for attr in tag_attrs:
				if 'src' in tag_attrs:
					src = tag_attrs[attr]
					urlib = self.framework.urlib(src)
					if urlib.check_urlfile('js'):
						resp.append(src)
		return resp

	@property
	def get_cssfiles(self):
		self.remove_comments
		reg = r"<(?i)link[^>]+/>"
		reg = re.compile(reg)
		resp = []
		find = reg.findall(self.page)
		for tag in find:
			tag_attr = self.get_attrs(tag, ['href'])
			for link in tag_attr:
				if 'href' in link:
					href = link.get('href')
					urlib = self.framework.urlib(href)
					if urlib.check_urlfile('css'):
						resp.append(href)
		return resp

	@property
	def get_links(self):
		self.remove_comments
		reg = r'[\'"](/.*?)[\'"]|[\'"](http.*?)[\'"]'
		reg = re.compile(reg)
		find = reg.findall(self.page)
		links = []
		for link in find:
			link = list(link)
			link.pop(link.index(''))
			link = link[0]
			if not re.search(r'<|>|/>', link):
				link = link.replace('\'', '').replace('"', '')
				links.append(link)
		return links

	@property
	def get_ahref(self):
		self.remove_comments
		reg = re.compile(r'<[aA].*(href|HREF)=([^\s>]+)')
		find = reg.findall(self.page)
		links = []
		for link in find:
			link = list(link)
			link.pop(link.index(''))
			link = link[0]
			links.append(link)
		return links

	@property
	def get_credit_cards(self):
		reg = re.compile(r"[0-9]{4}[ ]?[-]?[0-9]{4}[ ]?[-]?[0-9]{4}[ ]?[-]?[0-9]{4}")
		find = reg.findall(self.page)
		return list(set(find))

	@property
	def get_html_comments(self):
		reg = re.compile(r'<!--(.*?)-->')
		js_reg = re.compile(r'/\*(.*?)\*/')
		find = reg.findall(self.page)
		find.extend(js_reg.findall(self.page))
		return list(set(find))

	@property
	def get_forms(self):
		resp = {}
		self.remove_comments
		reg = re.compile(r'(?i)(?s)<form.*?</form.*?>')
		forms = reg.findall(self.page)
		form_attrs = ['action', 'method', 'name', 'autocomplete', 'novalidate', 'target', 'role']
		input_attrs = ['accept', 'alt', 'disabled', 'form', 'formaction', 'formenctype', 'formmethod',
					   'max', 'maxlength', 'min', 'minlength', 'name', 'pattern', 'readonly', 'required',
					   'selected', 'size', 'src', 'type', 'value', 'for']
		for form in range(len(forms)):
			form_tag = re.compile(r'<(?i)form.*?>')
			form_tag = form_tag.findall(forms[form])[0]
			get_form_attrs = self.get_attrs(form_tag, form_attrs)
			if 'action' not in get_form_attrs:
				get_form_attrs['action'] = '/'
			resp[form] = {'form': get_form_attrs}
			resp[form]['inputs'] = {}
			inputs = re.findall(r'<(?i)input.*?/?>', forms[form])
			textareas = re.findall(r'<textarea.*?>', forms[form])

			for inp in range(len(textareas)):
				inp_attrs = self.get_attrs(textareas[inp], input_attrs)
				if 'textarea' in resp[form]['inputs']:
					resp[form]['inputs']['textarea'].append(inp_attrs)
				else:
					resp[form]['inputs'].update({'textarea': [inp_attrs]})

			for inp in range(len(inputs)):
				inp_attrs = self.get_attrs(inputs[inp], input_attrs)
				if not 'type' in inp_attrs:
					type_attr = 'text'
					inp_attrs['type'] = 'text'

				type_attr = inp_attrs['type'].lower()
				# Set default submit value
				if type_attr == 'submit' and 'value' not in inp_attrs:
					inp_attrs['value'] = 'Submit Query'

				if type_attr in resp[form]['inputs']:
					resp[form]['inputs'][type_attr].append(inp_attrs)
				else:
					resp[form]['inputs'].update({type_attr: [inp_attrs]})

		return resp



"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

class main:

	def __init__(self, q, limit=1):
		""" metacrawler.com search engine

			q 		  : query for search
			limit	  : count of pages
		"""
		self.framework = main.framework
		self.q = self.framework.urlib(q).quote
		self.xpath_name = {
			'results': '//div[@class="web-bing"]',
			'results_content': './/span[@class="web-bing__description"]',
			'results_title_and_a': './/a[@class="web-bing__title"]',
			'results_cite': './/span[@class="web-bing__url"]'
		}
		self.xpath = {
			self.xpath_name['results']: [
				self.xpath_name['results_content'],
				self.xpath_name['results_title_and_a'],
				self.xpath_name['results_cite']
			]
		}
		self.limit = limit
		self._pages = ''
		self.metacrawler = 'https://www.metacrawler.com/serp'

	def run_crawl(self):
		page = 1
		payload = {'q': self.q, 'page': 1}
		headers = {
			'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) \
					AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.1 Safari/605.1.15',
			'Accept-Language': 'en-US,en;q=0.5'
		}

		while True:
			self.framework.verbose(f"[METACRAWLER] Searching in {page} page...")
			try:
				req = self.framework.request(url=self.metacrawler, params=payload, headers=headers)
			except Exception as err:
				self.framework.error('ConnectionError', 'util/engines/metacrawler', 'run_crawl')
				self.framework.error('Metacrawler is missed!', 'util/engines/metacrawler', 'run_crawl')
				break
			else:
				text = self.framework.to_str(req.text)
				if 'To continue, please respond below:' in text:
					self.framework.error('CaptchaError', 'util/engines/metacrawler', 'run_crawl')
					self.framework.error('Metacrawler is missed!', 'util/engines/metacrawler', 'run_crawl')
					break
				self._pages += text
				if 'Next »' not in text or page >= self.limit:
					break
				page += 1
				payload['page'] += 1

	@property
	def pages(self):
		return self._pages

	@property
	def links(self):
		links = self.framework.page_parse(self._pages).findall(\
			r'<a class="web-bing__title" href="(.*)"\sdata-thash')
		return links

	@property
	def results(self):
		parser = self.framework.page_parse(self._pages)
		xpath_results = parser.html_fromstring(self.xpath)
		results = []
		if xpath_results:
			root = xpath_results[self.xpath_name['results']]
			for i in range(len(root[self.xpath_name['results_title_and_a']])):
				t_and_a = root[self.xpath_name['results_title_and_a']][i]
				a = t_and_a.get('href')
				t = t_and_a.text_content()
				result = {
					't': t,
					'a': a,
					'c': root[self.xpath_name['results_cite']][i].text_content(),
					'd': root[self.xpath_name['results_content']][i].text_content(),
				}
				results.append(result)
		return results

	@property
	def dns(self):
		return self.framework.page_parse(self._pages).get_dns(self.q)
	
	@property
	def emails(self):
		return self.framework.page_parse(self._pages).get_emails(self.q)
	
	@property
	def docs(self):
		return self.framework.page_parse(self._pages).get_docs(self.q, self.links)



"""
OWASP Maryam!
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

meta = {
	'name': 'creativecommons',
	'author': 'Rishabh Jain',
	'version': '0.1',
	'description': 'Creative Commons is an American non-profit organization and international network devoted to educational access and expanding the range of creative works available for others to build upon legally and to share.',
	'sources': ('creativecommons',),
	'options': (
		('query', None, True, 'Query string', '-q', 'store', str),
		('limit', 15, False, 'Max result count (default=15)', '-l', 'store', int),
	),
	'examples': ('creativecommons -q <QUERY> -l 15 --output',)
}

def module_api(self):
    query = self.options['query']
    limit = self.options['limit']
    run = self.creativecommons(query, limit)
    run.run_crawl()
    output = {}
    count = run.total_result_found
    data = run.collected_data
    if count:
        output['Total Results'] = count 

    output['results'] = []
    
    if data:
        output['results'].extend(data)
    
    self.save_gather(output, 'search/creativecommons', query, output=self.options['output'])
    return output
    
def module_run(self):
    self.alert_results(module_api(self))


"""
OWASP Maryam!
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


class main :
    def __init__(self, q, limit=15):
        """ search.creativecommons.org search
        
            q 		  : The query for search
            limit	  : The number of details min 15 if exist
        """
        self.q = q
        self.limit = limit
        self.url = 'https://api.creativecommons.engineering/v1/images'
        self.params = {'q': self.q}
        self._collected_data = []
        self._result_found = 0
        
    def run_crawl(self):
        self.framework.verbose('[creativecommons] Extracting Data From API')
        
        # first page
        success, response = self.send_request(params=self.params)
        if not success: 
            return

        self._result_found = response['result_count']
        self.extract_data(response['results']) 
        total_pages = response['page_count']
        page_size = response['page_size']

        # all next pages
        if total_pages > 1 and len(self._collected_data) < self.limit:
            self.params['shouldPersistImages'] = 'true' 
            for i in range(self.limit//page_size + 1):
                self.framework.verbose(f"[creativecommons] Extracting Data From API request-{i+1}")
                success, response = self.send_request(params={**self.params, 'page': i+1})
                if not success: 
                    continue
                self.extract_data(response['results'])
        
    def send_request(self, params):
        try:
            response = self.framework.request(self.url, params=params)
        except:
            self.framework.error('Request Fail', 'util/creativecommons', 'send_request')
        else:
            if response.status_code != 200:
                self.framework.error('Request Fail - Invalid request', 'util/creativecommons', 'send_request')
                return False, None
            return True, response.json()

    def extract_data(self, data):
        for result in data:
            self._collected_data.append({
                'Title': result['title'],
                'Creator': result['creator'],
                'Creator_url': result['creator_url'],
                'Image-URL': result['url'],
            })

    @property
    def collected_data(self):
        return self._collected_data

    @property
    def total_result_found(self):
        return self._result_found


"""
OWASP Maryam!
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import re
import concurrent.futures
import json

meta = {
	'name': 'GitHub Leaks',
	'author': 'Aman Singh',
	'version': '0.1',
	'description': 'Search your query in the GitHub and show the potentially leaked info.',
	'sources': ('google', 'carrot2', 'bing','yahoo', 'millionshort', 'qwant', 'duckduckgo', 'github'),
	'options': (
		('query', None, True, 'Query string', '-q', 'store', str),
		('limit', 1, False, 'Search limit(number of pages, default=1)', '-l', 'store', int),
		('count', 50, False, 'Number of links per page(min=10, max=100, default=50)', '-c', 'store', int),
		('engine', 'google,github', False, 'Engine names for search(default=google, github)', '-e', 'store', str),
		('thread', 2, False, 'The number of engine that run per round(default=2)', '-t', 'store', int),
	),
    'examples': ('github_leaks-q <QUERY> -l 15 -e carrot2,bing,qwant --output',)
}

LINKS = []
PAGES = ''
REPO = []
OUTPUT = {'links': {}}

def checks(self, repo):
	checks = ['filename:.npmrc filename:.dockercfg filename:id_rsa filename:credentials filename:.s3cfg',#NOT MORE THAN 500 CHAR IN ONE QUERY
	'filename:wp-config.php filename:.htpasswd filename:.git-credentials filename:.bashrc filename:.bash_profile', 
	'filename:.netrc filename:config filename:connections.xml filename:express.conf filename:.pgpass', 
	'filename:proftpdpasswd filename:server.cfg filename:.bash_history filename:.cshrc filename:.history filename:.sh_history filename:sshd_config', 
	'filename:dhcpd.conf filename:prod.exs filename:shadow filename:passwd filename:.esmtprc filename:logins.json', 
	'filename:CCCam.cfg filename:settings filename:secrets filename:master.key filename:WebServers.xml',
	'password', 'DB_USERNAME', 'TOKEN', 'API_KEY' , 'SECRET_KEY', 'credentials',
	'extension:pem extension:ppk extension:sql extension:sls'
	]
	for term in checks:
		threadpool = concurrent.futures.ThreadPoolExecutor(max_workers=4)
		futures = (threadpool.submit(leaks, self, link, term) for link in repo)
		for results in concurrent.futures.as_completed(futures):
			print(f"Found {len(OUTPUT['links'])} links" , end= '\r')
	print('\n')

def leaks(self, repo, term):
	from bs4 import BeautifulSoup

	url = f"{repo}/search?q={self.urlib(term).quote_plus.replace('.','%2E')}"
	try:
		headers = {'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.102 Safari/537.36'}
		req = self.request(url, headers=headers, timeout=20)
		if 'We couldn’t find any code matching' in req.text:
			return
	except Exception as e:
		return
	else:
		try:
			soup = BeautifulSoup(req.content, 'html.parser')
			results = soup.find_all('div', class_='width-full')
			if len(results) > 0:
				for result in results:
					OUTPUT['links'][json.loads(result.find('a')['data-hydro-click'])['payload']['result']['url']] = [re.sub(' +', ' ', line) for line in str(''.join([code.text for code in result.find_all('td' ,class_ ="blob-code blob-code-inner")])).split('\n') if term in line]
		except:
			return

def search(self, name, q, q_formats, limit, count):
	global PAGES,LINKS, USERS, REPO
	engine = getattr(self, name)
	eng = name
	q = q_formats[f"{name}"] if f"{name}" in q_formats else q_formats['default']
	varnames = engine.__init__.__code__.co_varnames
	if 'limit' in varnames and 'count' in varnames:
		attr = engine(q, limit, count)
	elif 'limit' in varnames:
		attr = engine(q, limit)
	else:
		attr = engine(q)
	if eng == 'github':
		run = self.github(q)
		run.run_crawl()
		REPO += run.repositories
	else:
		attr.run_crawl()
		LINKS += attr.links
		PAGES += attr.pages


def module_api(self):
	global REPO
	query = self.options['query']
	limit = self.options['limit']
	count = self.options['count']
	engines = self.options['engine'].split(',')
	q_formats = {
		'default': f"site:github.com {query}",
		'millionshort': f'site:github.com "{query}"',
		'qwant': f'site:github.com {query}',
		'github': f'{query}'
	}

	self.thread(search, self.options['thread'], engines, query, q_formats, limit, count, meta['sources'])

	explore = self.page_parse(PAGES).get_networks
	bactery = ['marketplace', 'pulls', 'explore', 'issues','notifications', 'account', 'settings', 'login','about', 'pricing', 'site']

	for link in self.reglib().filter(r"https://(www\.)?github\.com/[\w-]{1,39}/[\w\-\.]+", list(set(LINKS))):
		repo = re.search(r"(https://(www\.)?github\.com/([\w-]{1,39})/[\w\-\.]+)", link)
		if repo.group(3) not in bactery:
			repo = repo.group(0)
			REPO.append(repo)

	REPO = list(set(REPO))
	checks(self, REPO)
	self.save_gather(OUTPUT,
	 'osint/github_leaks', query, output=self.options.get('output'))
	return OUTPUT

def module_run(self):
	output = module_api(self)
	self.alert('Links\n')
	for url in output['links']:
		self.output(url, 'G')
		for line in output['links'][url]:
			self.output(f"{line}")
		print('')



"""
OWASP Maryam!
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import re

meta = {
	'name': 'Common Vulnerabilities and Exposures Searcher',
	'author': 'Dimitrios Papageorgiou',
	'version': '0.3',
	'description': 'Search in open-sources to find CVEs(exploits, vulns).',
	'sources': ('mitre', 'nist', 'packetstormsecurity'),
	'options': (
			('query', None,
			 True, 'Query string', '-q', 'store', str),
			('engines', 'mitre', False,
			 'DB source to search. e.g mitre,...(default=mitre)', '-e', 'store', str),
			('count', 20, False,
			 'Number of results per search(default=20, -1 for all available)', '-c', 'store', int),
			('thread', 2, False,
			 'The number of engine that run per round(default=2)', '-t', 'store', int),
	),
	'examples': ('cve_search -q sql -e nist --output',)
}

TITLES = []
LINKS = []
DESCRIPTIONS = []

def set_data(titles, links, descs):
	global TITLES,LINKS,DESCRIPTIONS
	TITLES += titles
	LINKS += links
	DESCRIPTIONS += descs

def search(self, source, q, q_formats, limit, count):
	engine = eval(source)(self, q, count)

def mitre(self, q, count):
	self.verbose('[MITRE] Searching in mitre...')
	try:
		req = self.request(
			f"https://cve.mitre.org/cgi-bin/cvekey.cgi?keyword={q}")
	except Exception as e:
		self.error('Mitre is missed!', 'cve_search', 'mitre')
	else:
		hrefs = re.findall(
			r'<a href="(/cgi-bin/cvename.cgi\?name=[^"]+)">', req.text)[:count]
		titles = re.findall(
			r'<a href="[^"]+">(CVE-[^<]+)</a>', req.text)[:count]
		links = list(map(lambda x: f"https://cve.mitre.org{x}", hrefs))
		desc = re.findall(
			r'<td valign="top">(.*?)<\/td>', req.text.replace('\n', ''))[:-1][:count]
		set_data(titles, links, desc)

def nist(self, q, count):
	self.verbose('[NIST] Searching in nist...')
	try:
		req = self.request(
			f"https://services.nvd.nist.gov/rest/json/cves/1.0?keyword={q}&resultsPerPage={count}")
	except Exception as e:
		self.error('Nist is missed!', 'cve_search', 'nist')
	else:
		cve_items = req.json()['result']['CVE_Items']
		titles = [cve['cve']['CVE_data_meta']['ID'] for cve in cve_items]
		links = [f"https://nvd.nist.gov/vuln/detail/{id}" for id in titles]
		desc = [cve['cve']['description']['description_data'][0]['value']
				for cve in cve_items]
		set_data(titles, links, desc)

def packetstormsecurity(self, q, count):
	self.verbose('[Packetstormsecurity] Searching in packetstormsecurity...')
	try:
		req = self.request(
			f"https://packetstormsecurity.com/search/?q={q}", timeout=60)
	except Exception as e:
		self.error('Packetstormsecurity is missed', 'cve_search', 'packetstormsecurity')
	else:
		titles = re.findall(r'<a class="ico text-plain"[^>]+>([^<]+)</a>', req.text)[:count]
		links = re.findall(r'<a class="ico text-plain" href="([^"]+)"[^>]+>', req.text)[:count]
		links = [f"https://packetstormsecurity.com{link}" for link in links]
		desc = re.findall(r'<dd class="detail"><p>([^<]+)</p></dd>', req.text)[:count]
		set_data(titles, links, desc)

def module_api(self):
	query = self.options['query']
	engines = self.options['engines'].split(',')
	count = self.options['count']
	output = {'exploits': []}
	self.thread(search, self.options['thread'], engines, query, {}, 3, count, meta['sources'])
	output['exploits'] = [[TITLES[x], LINKS[x], DESCRIPTIONS[x]] for x in range(len(TITLES))]
	self.save_gather(output, 'osint/cve_search', query, output=self.options.get('output'))
	return output

def module_run(self):
	output = module_api(self)
	for item in output['exploits']:
		self.output(item[0], 'G')
		self.output(item[1], 'R')
		self.output(f"{item[2]}\n")



"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import lxml

class main:

	def __init__(self, q, limit=15):
		""" scholar.google.com search engine

			q	  : query for search
			limit	  : maximum result count
		"""
		self.framework = main.framework
		self.q = q
		self.max = limit
		self._articles = []
		self._results = []

	def run_crawl(self):
		start = 0
		self.framework.verbose('Searching the scholar.google.com domain...')
		for start in range(0, (self.max//10)+1):
			url = f"https://scholar.google.com/scholar"
			payload = {'as_vis': '1', 'q': self.q, 'start': start}
			try:
				req = self.framework.request(
						url=url,
						params=payload)
			except:
				self.framework.error('[SCHOLAR] ConnectionError', 'util/search/scholar', 'run_crawl')
				self.framework.error('Google Scholar is missed!', 'util/search/scholar', 'run_crawl')
			else:
				doc = lxml.html.document_fromstring(req.text)
				self._articles.extend(doc.find_class('gs_ri'))
				start += 10
	
	@property
	def articles(self):
		return self._articles

	@property
	def results(self):
		findlink = lambda x: x.xpath('h3/a')[0].attrib['href']
		findauthors = lambda x: x.find_class('gs_a')[0].text_content()
		findtitle = lambda x: x.xpath('h3/a')[0].text_content()
		finddesc = lambda x: x.find_class('gs_rs')[0].text_content()

		for count,article in enumerate(self._articles):
			if count==self.max:
				break
			self._results.append({
				't': findtitle(article),
				'a': findlink(article),
				'c': findauthors(article),
				'd': finddesc(article)
				})

		return self._results



"""
OWASP Maryam!

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import re

class main:

	def __init__(self, string=None):
		self.string = str(string)

	protocol_s = r"^([A-z0-9]+:\/\/)"
	protocol_m = r"([A-z0-9]+:\/\/)"
	email_s = r"^\w+@[A-z_\-.0-9]{5,255}$"
	email_m = r"\w+@[A-z_\-.0-9]{5,255}"
	phone_s = r"^([0-9]( |-)?)?(\(?[0-9]{3}\)?|[0-9]{3})( |-)?([0-9]{3}( |-)?[0-9]{4}|[A-z0-9]{7})$"
	phone_m = r"([0-9]( |-)?)?(\(?[0-9]{3}\)?|[0-9]{3})( |-)?([0-9]{3}( |-)?[0-9]{4}|[A-z0-9]{7})"
	domain_s = r"^([A-z0-9]([A-z0-9\-]{0,61}[A-z0-9])?\.)+[A-z]{2,6}(\:[0-9]{1,5})*$"
	domain_m = r"[A-z0-9\-]{0,61}\.+[A-z]{2,6}"
	url_s = r"^([A-z0-9]+:\/\/)?(www.|[A-z0-9].)[A-z0-9\-\.]+\.[A-z]{2,6}(\:[0-9]{1,5})*(\/($|[A-z0-9.,;?\'\\+&amp;%$#=~_-]+))*$"
	url_m = r"ftp|https?://[A-z0-9\-.]{2,255}[\/A-z\.:\-0-9%~@#?&()+_;,\']+"
	id_s = r"^@[A-z_0-9\.\-]{2,255}$"
	id_m = r"@[A-z_0-9\.\-]{2,255}"
	ip_s = r"^\d+\.[\d]+\.[\d]+\.[\d]+$"
	ip_m = r"\d+\.[\d]+\.[\d]+\.[\d]+"
	social_network_ulinks = {
		'Instagram': r"instagram\.com/[A-z_0-9.\-]{1,30}",
		'Facebook': [r"facebook\.com/[A-z_0-9\-]{2,50}", r"fb\.com/[A-z_0-9\-]{2,50}"],
		'Twitter': r"twitter\.com/[A-z_0-9\-.]{2,40}",
		'Github': r"github\.com/[A-z0-9_-]{1,39}",
		'Github site': [r"[A-z0-9_-]{1,39}\.github\.io", r"[A-z0-9_-]{1,39}\.github\.com"],
		'Telegram': r"telegram\.me/[A-z_0-9]{5,32}",
		'Youtube user': r"youtube\.com/user/[A-z_0-9\-\.]{2,100}",
		'Youtube channel': [r"youtube\.com/c/[A-z_0-9\-\.]{2,100}", \
				r"youtube\.com/channel/[A-z_0-9\-\.]{2,100}"],
		'Linkedin company': r"linkedin\.com/company/[A-z_0-9\.\-]{3,50}",
		'Linkedin individual': r"linkedin\.com/in/[A-z_0-9\.\-]{3,50}",
		'Googleplus': r"\.?plus\.google\.com/[A-z0-9_\-.+]{3,255}",
		'WordPress': r"[A-z0-9\-]+\.wordpress\.com",
		'Reddit': r"reddit\.com/user/[A-z0-9_\-]{3,20}",
		'Tumblr': r"[A-z0-9\-]{3,32}\.tumblr\.com",
		'Blogger': r"[A-z0-9\-]{3,50}\.blogspot\.com"
		}

	def search(self, regex, _type=list):
		regex = re.compile(regex)
		regex = regex.findall(self.string)
		return regex

	def sub(self, regex, sub_string):
		data = re.sub(regex, sub_string, self.string)
		return data
	
	def filter(self, regex, _list: list) -> list:
		if not isinstance(regex, str):
			return filter(regex, _list)
		else:
			return filter(re.compile(regex).match, _list)

	@property
	def emails(self):
		emails = self.search(self.email_m)
		return emails
	
	@property
	def urls(self):
		urls = self.search(self.url_m)
		return urls
	
	@property
	def domains(self):
		domains = self.search(self.domain_m)
		return domains
